from __future__ import annotations

from types import SimpleNamespace

from gotta.plugins import github


def test_parse_args_supports_status_subcommand() -> None:
    assert github.parse_args(["status"]) == github.ParsedArgs(command="status", output="summary")
    assert github.parse_args(["status", "--output", "summary"]) == github.ParsedArgs(
        command="status",
        output="summary",
    )


def test_parse_args_defaults_render_to_markdown() -> None:
    assert github.parse_args(["https://github.com/acme/widgets"]) == github.ParsedArgs(
        command="render",
        output="markdown",
        url="https://github.com/acme/widgets",
    )
    assert github.parse_args(
        ["--output", "markdown", "https://github.com/acme/widgets"]
    ) == github.ParsedArgs(
        command="render",
        output="markdown",
        url="https://github.com/acme/widgets",
    )
    assert github.parse_args(
        ["https://github.com/acme/widgets", "--output", "markdown"]
    ) == github.ParsedArgs(
        command="render",
        output="markdown",
        url="https://github.com/acme/widgets",
    )


def test_github_canonical_locator_and_preferred_name_ignore_flag_order() -> None:
    argv = ["https://github.com/acme/widgets", "--output", "summary"]
    options = SimpleNamespace(save_as="")

    assert github.canonical_locator(argv) == "https://github.com/acme/widgets"
    assert github.preferred_name(argv, options) == "widgets.summary"


def test_parse_args_supports_search_surface() -> None:
    assert github.parse_args(["search", "relay"]) == github.ParsedArgs(
        command="search",
        output="markdown",
        query="relay",
        search_type="repo",
        repo="",
        limit=10,
    )
    assert github.parse_args(["search", "--global", "ABC"]) == github.ParsedArgs(
        command="search",
        output="markdown",
        query="ABC",
        search_type="repo",
        repo="",
        limit=10,
        global_search=True,
    )
    assert github.parse_args(
        ["search", "--type", "pr", "--repo", "acme/widgets", "--limit", "5", "ABC", "proxy"]
    ) == github.ParsedArgs(
        command="search",
        output="markdown",
        query="ABC proxy",
        search_type="pr",
        repo="acme/widgets",
        limit=5,
    )
    assert github.parse_args(
        [
            "search",
            "--type",
            "code",
            "--repo",
            "acme/widgets",
            "--filename",
            "package.json",
            "--extension",
            "json",
            "--match",
            "path",
            "lint",
        ]
    ) == github.ParsedArgs(
        command="search",
        output="markdown",
        query="lint",
        search_type="code",
        repo="acme/widgets",
        filename="package.json",
        extension="json",
        match="path",
        limit=10,
    )


def test_parse_args_supports_search_help(capsys) -> None:
    assert github.main(["search", "--help"]) == 0
    output = capsys.readouterr().out
    assert "usage: gotta github search" in output
    assert "--global" in output


def test_github_help_renders_once(capsys) -> None:
    assert github.main(["--help"]) == 0
    output = capsys.readouterr().out
    assert output.count("usage: gotta github [status") == 1
    assert "/commits/HEAD" in output


def test_github_search_canonical_locator_and_preferred_name() -> None:
    options = SimpleNamespace(save_as="")

    argv = ["search", "--type", "pr", "--repo", "acme/widgets", "ABC", "proxy"]

    assert github.canonical_locator(argv) == "github:search --type pr --repo acme/widgets ABC proxy"
    assert github.preferred_name(argv, options) == "github-search-prs-acme-widgets-abc-proxy.md"
    assert github.canonical_locator(["search", "--global", "ABC"]) == "github:search --global ABC"
    assert (
        github.preferred_name(["search", "--global", "ABC"], options)
        == "github-search-repos-global-abc.md"
    )
    code_argv = [
        "search",
        "--type",
        "code",
        "--repo",
        "acme/widgets",
        "--filename",
        "package.json",
        "lint",
    ]
    assert (
        github.canonical_locator(code_argv)
        == "github:search --type code --repo acme/widgets --filename package.json lint"
    )
    assert (
        github.preferred_name(code_argv, options)
        == "github-search-code-acme-widgets-file-package.json-lint.md"
    )


def test_github_preferred_name_is_specific_for_rendered_object_urls() -> None:
    options = SimpleNamespace(save_as="")

    assert (
        github.preferred_name(["https://github.com/acme/widgets/pull/19"], options)
        == "widgets-pr-19.md"
    )
    assert (
        github.preferred_name(["https://github.com/acme/widgets/issues/19"], options)
        == "widgets-issue-19.md"
    )
    assert (
        github.preferred_name(["https://github.com/acme/widgets/commits/main"], options)
        == "widgets-commits-main.md"
    )
    assert (
        github.preferred_name(["https://github.com/acme/widgets/commits"], options)
        == "widgets-commits.md"
    )


def test_route_target_accepts_clean_supported_github_urls() -> None:
    url = "https://github.com/acme/widgets/commits/main"

    assert github.route_target(url) == [url]
    assert github.route_target(f"{url}#history") == [f"{url}#history"]
    assert github.route_target("github:github.com/acme/widgets/commits/main") == [url]


def test_route_target_strips_fragments_from_supported_tree_urls() -> None:
    url = "https://github.com/acme/widgets/tree/main/docs"

    assert github.route_target(f"{url}#readme") == [f"{url}#readme"]


def test_route_target_rejects_whitespace_contaminated_github_urls() -> None:
    url = "https://github.com/acme/widgets/commits/main"

    assert github.route_target(f"{url} --limit 20") is None
    assert github.route_target("github:github.com/acme/widgets/commits/main --limit 20") is None


def test_main_supports_commit_urls(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(
        github,
        "gh_json_object",
        lambda gh, args: {
            "sha": "abcdef1",
            "html_url": "https://github.com/acme/widgets/commit/abcdef1",
            "commit": {
                "message": "Tighten control surface",
                "author": {"name": "Alice", "date": "2026-03-12T12:00:00Z"},
            },
            "files": [
                {
                    "filename": "README.md",
                    "status": "modified",
                    "additions": 3,
                    "deletions": 1,
                    "changes": 4,
                }
            ],
        },
    )

    assert github.main(["https://github.com/acme/widgets/commit/abcdef1"]) == 0
    output = capsys.readouterr().out
    assert "# acme/widgets commit abcdef1" in output
    assert "## Files" in output


def test_main_supports_commit_history_urls(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(
        github,
        "gh_json_value",
        lambda gh, args: [
            {
                "sha": "abcdef123456",
                "html_url": "https://github.com/acme/widgets/commit/abcdef123456",
                "commit": {
                    "message": "Add continuity model\n\nDetails",
                    "author": {"name": "Alice", "date": "2026-03-12T12:00:00Z"},
                },
            }
        ],
    )

    assert github.main(["https://github.com/acme/widgets/commits/main"]) == 0
    output = capsys.readouterr().out
    assert "# acme/widgets commit history for `main`" in output
    assert "- **Created:** 2026-03-12T12:00:00Z" in output
    assert "- **Updated:** 2026-03-12T12:00:00Z" in output
    assert "Add continuity model" in output


def test_main_supports_commit_history_limit_for_url_renders(monkeypatch, capsys) -> None:
    seen: list[list[str]] = []

    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)

    def fake_gh_json_value(gh, args):
        seen.append(args)
        return [
            {
                "sha": "abcdef123456",
                "html_url": "https://github.com/acme/widgets/commit/abcdef123456",
                "commit": {
                    "message": "Add continuity model\n\nDetails",
                    "author": {"name": "Alice", "date": "2026-03-12T12:00:00Z"},
                },
            }
        ]

    monkeypatch.setattr(github, "gh_json_value", fake_gh_json_value)

    assert github.main(["https://github.com/acme/widgets/commits/HEAD", "--limit", "50"]) == 0
    assert "&per_page=50" in seen[0][1]
    output = capsys.readouterr().out
    assert "# acme/widgets commit history for `HEAD`" in output


def test_main_rejects_limit_for_non_list_url_shapes(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)

    assert github.main(["https://github.com/acme/widgets/issues/19", "--limit", "10"]) == 2
    err = capsys.readouterr().err
    assert "`--limit` is only supported for GitHub commit-history URLs" in err
    assert "/commits/HEAD" in err


def test_main_supports_commit_history_root_urls(monkeypatch, capsys) -> None:
    seen: list[list[str]] = []

    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)

    def fake_gh_json_object(gh, args):
        seen.append(args)
        return {"default_branch": "main"}

    monkeypatch.setattr(github, "gh_json_object", fake_gh_json_object)
    monkeypatch.setattr(
        github,
        "gh_json_value",
        lambda gh, args: [
            {
                "sha": "abcdef123456",
                "html_url": "https://github.com/acme/widgets/commit/abcdef123456",
                "commit": {
                    "message": "Add continuity model\n\nDetails",
                    "author": {"name": "Alice", "date": "2026-03-12T12:00:00Z"},
                },
            }
        ],
    )

    assert github.main(["https://github.com/acme/widgets/commits"]) == 0
    output = capsys.readouterr().out
    assert seen == [["api", "repos/acme/widgets"]]
    assert "# acme/widgets commit history for `main`" in output


def test_main_supports_path_scoped_commit_history_urls(monkeypatch, capsys) -> None:
    seen: list[list[str]] = []

    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)

    def fake_gh_json_value(gh, args):
        seen.append(args)
        return [
            {
                "sha": "abcdef123456",
                "html_url": "https://github.com/acme/widgets/commit/abcdef123456",
                "commit": {
                    "message": "Add continuity model\n\nDetails",
                    "author": {"name": "Alice", "date": "2026-03-12T12:00:00Z"},
                },
            }
        ]

    monkeypatch.setattr(github, "gh_json_value", fake_gh_json_value)

    assert github.main(["https://github.com/acme/widgets/commits/main/docs/adr"]) == 0
    output = capsys.readouterr().out
    assert "&path=docs/adr" in seen[0][1]
    assert "- **Path:** `docs/adr`" in output
    assert "- **URL:** https://github.com/acme/widgets/commits/main/docs/adr" in output


def test_main_commit_history_invalid_ref_suggests_default_branch(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(
        github,
        "gh_json_value",
        lambda gh, args: (_ for _ in ()).throw(RuntimeError("gh: Not Found (HTTP 404)")),
    )
    monkeypatch.setattr(github, "default_branch_name", lambda gh, owner, repo: "trunk")

    assert github.main(["https://github.com/acme/widgets/commits/main"]) == 1
    err = capsys.readouterr().err
    assert "commit-history ref `main` was not found" in err
    assert "Specific `/commits/<ref>` URLs require a real branch, tag, or commit-ish." in err
    assert "https://github.com/acme/widgets/commits" in err
    assert "https://github.com/acme/widgets/commits/HEAD" in err
    assert "https://github.com/acme/widgets/commits/trunk" in err


def test_main_supports_repo_search(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(github, "_accessible_owner_targets", lambda gh: [("org", "acme")])

    def fake_gh_json_object(gh, args):
        assert "search/repositories" in args[1]
        if "org:acme" in args[1]:
            return {
                "total_count": 1,
                "items": [
                    {
                        "full_name": "acme/relay-core",
                        "name": "relay-core",
                        "html_url": "https://github.com/acme/relay-core",
                        "description": "WireGuard overlay",
                        "language": "Go",
                        "stargazers_count": 17,
                        "created_at": "2026-01-01T10:00:00Z",
                        "updated_at": "2026-03-10T12:00:00Z",
                        "pushed_at": "2026-03-11T09:30:00Z",
                        "default_branch": "main",
                        "owner": {"login": "acme"},
                    }
                ],
            }
        raise AssertionError("default owned-scope search should not hit the global corpus")

    monkeypatch.setattr(github, "gh_json_object", fake_gh_json_object)

    assert github.main(["search", "relay"]) == 0
    output = capsys.readouterr().out
    assert "### GitHub Search: relay" in output
    assert "- _Search scope_: owned repositories and visible organizations only" in output
    assert "- Created:" in output
    assert "- Updated: 2026-03-11T09:30:00Z" in output
    assert "[acme/relay-core](https://github.com/acme/relay-core)" in output
    assert "[public/noise]" not in output


def test_main_supports_global_repo_search_excluding_owned_results(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(github, "_accessible_owner_targets", lambda gh: [("org", "acme")])

    def fake_gh_json_object(gh, args):
        assert "search/repositories" in args[1]
        return {
            "total_count": 3,
            "items": [
                {
                    "full_name": "acme/relay-core",
                    "name": "relay-core",
                    "html_url": "https://github.com/acme/relay-core",
                    "description": "Owned result",
                    "owner": {"login": "acme"},
                },
                {
                    "full_name": "public/noise",
                    "name": "noise",
                    "html_url": "https://github.com/public/noise",
                    "description": "Global noise",
                    "owner": {"login": "public"},
                },
                {
                    "full_name": "public/abc",
                    "name": "abc",
                    "html_url": "https://github.com/public/abc",
                    "description": "Global ABC noise",
                    "owner": {"login": "public"},
                },
            ],
        }

    monkeypatch.setattr(github, "gh_json_object", fake_gh_json_object)

    assert github.main(["search", "--global", "ABC"]) == 0
    output = capsys.readouterr().out
    assert "- _Search scope_: global GitHub excluding owned-scope results" in output
    assert "[acme/relay-core]" not in output
    assert "[public/noise](https://github.com/public/noise)" in output


def test_main_supports_pull_request_search(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(
        github,
        "gh_json_object",
        lambda gh, args: {
            "total_count": 1,
            "items": [
                {
                    "title": "Replace ABC gateway",
                    "number": 27,
                    "html_url": "https://github.com/acme/widgets/pull/27",
                    "state": "open",
                    "user": {"login": "alice"},
                    "repository_url": "https://api.github.com/repos/acme/widgets",
                    "created_at": "2026-03-01T10:00:00Z",
                    "updated_at": "2026-03-12T09:15:00Z",
                    "body": "This replaces ABC.",
                    "pull_request": {"url": "https://api.github.com/repos/acme/widgets/pulls/27"},
                    "labels": [{"name": "continuity"}],
                }
            ],
        },
    )

    assert github.main(["search", "--type", "pr", "--repo", "acme/widgets", "ABC"]) == 0
    output = capsys.readouterr().out
    assert "### GitHub Search: ABC" in output
    assert "- _Repo Scope_: `acme/widgets`" in output
    assert "- Created: 2026-03-01T10:00:00Z" in output
    assert "[acme/widgets pr #27: Replace ABC gateway](https://github.com/acme/widgets/pull/27)" in output


def test_main_supports_code_search(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(
        github,
        "search_code_payload",
        lambda gh, **kwargs: {
            "surface": "github",
            "type": "code",
            "query": "lint",
            "scopeRepo": "acme/widgets",
            "filename": "package.json",
            "extension": "",
            "language": "",
            "match": "",
            "searchPlan": "repo-scope",
            "accessibleOwners": [],
            "scopedResultCount": 0,
            "globalResultCount": 1,
            "resultCount": 1,
            "results": [
                {
                    "kind": "code",
                    "repository": "acme/widgets",
                    "path": "package.json",
                    "sha": "abcdef123456",
                    "url": "https://github.com/acme/widgets/blob/abcdef123456/package.json",
                    "textMatches": ['"lint": "npm run lint"'],
                }
            ],
        },
    )

    assert (
        github.main(
            [
                "search",
                "--type",
                "code",
                "--repo",
                "acme/widgets",
                "--filename",
                "package.json",
                "lint",
            ]
        )
        == 0
    )
    output = capsys.readouterr().out
    assert "- _Type_: `code`" in output
    assert "- _Filename_: `package.json`" in output
    assert "[acme/widgets:package.json](https://github.com/acme/widgets/blob/abcdef123456/package.json)" in output
    assert '"lint": "npm run lint"' in output


def test_search_payload_defaults_to_owned_scope(monkeypatch) -> None:
    monkeypatch.setattr(github, "_accessible_owner_targets", lambda gh: [("org", "acme")])

    def fake_gh_json_object(gh, args):
        if "org:acme" in args[1]:
            return {
                "total_count": 1,
                "items": [
                    {
                        "full_name": "acme/signal",
                        "name": "signal",
                        "html_url": "https://github.com/acme/signal",
                        "owner": {"login": "acme"},
                    }
                ],
            }
        raise AssertionError("default owned-scope search should not hit the global corpus")

    monkeypatch.setattr(github, "gh_json_object", fake_gh_json_object)

    payload = github.search_repositories_payload(
        "gh",
        query="ABC",
        repo="",
        limit=2,
        global_search=False,
    )

    assert payload["searchPlan"] == "owned-only"
    assert payload["results"][0]["fullName"] == "acme/signal"
    assert payload["scopedResultCount"] == 1
    assert payload["globalResultCount"] == 0


def test_search_payload_global_excludes_owned_hits(monkeypatch) -> None:
    monkeypatch.setattr(github, "_accessible_owner_targets", lambda gh: [("org", "acme")])
    monkeypatch.setattr(
        github,
        "gh_json_object",
        lambda gh, args: {
            "total_count": 3,
            "items": [
                {
                    "full_name": "acme/signal",
                    "name": "signal",
                    "html_url": "https://github.com/acme/signal",
                    "owner": {"login": "acme"},
                },
                {
                    "full_name": "public/noise",
                    "name": "noise",
                    "html_url": "https://github.com/public/noise",
                    "owner": {"login": "public"},
                },
                {
                    "full_name": "public/other-noise",
                    "name": "other-noise",
                    "html_url": "https://github.com/public/other-noise",
                    "owner": {"login": "public"},
                },
            ],
        },
    )

    payload = github.search_repositories_payload(
        "gh",
        query="ABC",
        repo="",
        limit=3,
        global_search=True,
    )

    assert payload["searchPlan"] == "global-excluding-owned"
    assert [item["fullName"] for item in payload["results"]] == [
        "public/noise",
        "public/other-noise",
    ]


def test_main_falls_back_to_directory_readme_for_readme_blob(monkeypatch) -> None:
    seen: dict[str, object] = {}

    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)

    def fake_fetch(gh, *, owner, repo, ref, path):
        raise RuntimeError("404")

    monkeypatch.setattr(github, "fetch_content_file", fake_fetch)
    monkeypatch.setattr(
        github,
        "load_directory_readme",
        lambda gh, *, owner, repo, ref, path="", entries=None: ("docs/README.md", b"# Docs\n"),
    )
    monkeypatch.setattr(
        github,
        "render_content",
        lambda data, path: seen.update({"data": data, "path": path}),
    )

    assert (
        github.main(
            ["--output", "markdown", "https://github.com/acme/widgets/blob/main/README.md"]
        )
        == 0
    )
    assert seen["path"] == "docs/README.md"
    assert seen["data"] == b"# Docs\n"


def test_main_repo_markdown_falls_back_to_directory_listing(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(
        github,
        "gh_json_object",
        lambda gh, args: {
            "name": "widgets",
            "visibility": "private",
            "url": "https://github.com/acme/widgets",
            "createdAt": "2026-03-01T10:00:00Z",
            "updatedAt": "2026-03-10T11:30:00Z",
            "pushedAt": "2026-03-11T09:15:00Z",
            "defaultBranchRef": {"name": "main"},
        },
    )
    monkeypatch.setattr(github, "load_directory_readme", lambda *args, **kwargs: None)
    monkeypatch.setattr(
        github,
        "list_directory_entries",
        lambda gh, *, owner, repo, ref, path="": [
            {"name": "README.md", "path": "README.md", "type": "file", "size": 12},
            {"name": "docs", "path": "docs", "type": "dir"},
        ],
    )

    assert github.main(["--output", "markdown", "https://github.com/acme/widgets"]) == 0
    output = capsys.readouterr().out
    assert "# widgets" in output
    assert "https://github.com/acme/widgets/tree/main" in output
    assert "- **README:** [README.md](https://github.com/acme/widgets/blob/main/README.md)" in output


def test_main_repo_markdown_uses_fragment_to_render_root_readme(monkeypatch) -> None:
    seen: dict[str, object] = {}

    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(
        github,
        "gh_json_object",
        lambda gh, args: {
            "name": "widgets",
            "visibility": "private",
            "url": "https://github.com/acme/widgets",
            "createdAt": "2026-03-01T10:00:00Z",
            "updatedAt": "2026-03-10T11:30:00Z",
            "pushedAt": "2026-03-11T09:15:00Z",
            "defaultBranchRef": {"name": "main"},
        },
    )
    monkeypatch.setattr(
        github,
        "list_directory_entries",
        lambda gh, *, owner, repo, ref, path="": [
            {"name": "README.md", "path": "README.md", "type": "file"},
            {"name": "docs", "path": "docs", "type": "dir"},
        ],
    )
    monkeypatch.setattr(
        github,
        "fetch_content_file",
        lambda gh, *, owner, repo, ref, path: {
            "type": "file",
            "encoding": "base64",
            "content": "IyBSb290IERvY3MK",
        },
    )
    monkeypatch.setattr(
        github,
        "render_content",
        lambda data, path: seen.update({"data": data, "path": path}),
    )

    assert github.main(["https://github.com/acme/widgets#readme"]) == 0
    assert seen["path"] == "README.md"
    assert seen["data"] == b"# Root Docs\n"


def test_main_tree_markdown_defaults_to_directory_listing_even_with_readme(monkeypatch, capsys) -> None:
    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(
        github,
        "list_directory_entries",
        lambda gh, *, owner, repo, ref, path="": [
            {"name": "README.md", "path": "docs/README.md", "type": "file"},
            {"name": "QUICKSTART.md", "path": "docs/QUICKSTART.md", "type": "file"},
        ],
    )

    assert github.main(["https://github.com/acme/widgets/tree/main/docs"]) == 0
    output = capsys.readouterr().out
    assert "# acme/widgets:docs" in output
    assert "- **README:** [README.md](https://github.com/acme/widgets/blob/main/docs/README.md)" in output
    assert "QUICKSTART.md" in output


def test_main_tree_markdown_uses_fragment_to_render_matching_document(monkeypatch) -> None:
    seen: dict[str, object] = {}

    monkeypatch.setattr(github, "ensure_gh", lambda: "gh")
    monkeypatch.setattr(github, "ensure_gh_auth", lambda gh: None)
    monkeypatch.setattr(
        github,
        "list_directory_entries",
        lambda gh, *, owner, repo, ref, path="": [
            {"name": "README.md", "path": "docs/README.md", "type": "file"},
            {"name": "QUICKSTART.md", "path": "docs/QUICKSTART.md", "type": "file"},
        ],
    )
    monkeypatch.setattr(
        github,
        "fetch_content_file",
        lambda gh, *, owner, repo, ref, path: {
            "type": "file",
            "encoding": "base64",
            "content": "IyBRdWlja3N0YXJ0CgpVc2UgaXQgZmlyc3QuCg==",
        },
    )
    monkeypatch.setattr(
        github,
        "render_content",
        lambda data, path: seen.update({"data": data, "path": path}),
    )

    assert github.main(["https://github.com/acme/widgets/tree/main/docs#quickstart"]) == 0
    assert seen["path"] == "docs/QUICKSTART.md"
    assert seen["data"] == b"# Quickstart\n\nUse it first.\n"


def test_markdown_repo_includes_source_times() -> None:
    rendered = github.markdown_repo(
        {
            "name": "widgets",
            "url": "https://github.com/acme/widgets",
            "visibility": "private",
            "createdAt": "2026-03-01T10:00:00Z",
            "updatedAt": "2026-03-10T11:30:00Z",
            "pushedAt": "2026-03-11T09:15:00Z",
            "defaultBranchRef": {"name": "main"},
        }
    )

    assert "- **Created:** 2026-03-01T10:00:00Z" in rendered
    assert "- **Updated:** 2026-03-10T11:30:00Z" in rendered
    assert "- **Pushed:** 2026-03-11T09:15:00Z" in rendered



from __future__ import annotations
import os
from pathlib import Path

import pytest

from gotta import content
from gotta import target
from gotta.plugins import peer, read, session


def test_read_local_directory_renders_native_listing(tmp_path: Path, capsys) -> None:
    root = tmp_path / "repo"
    root.mkdir()
    (root / ".git").mkdir()
    (root / ".git" / "config").write_text("[core]\n\trepositoryformatversion = 0\n", encoding="utf-8")
    (root / "docs").mkdir()
    (root / "docs" / "README.md").write_text("# Docs\n", encoding="utf-8")

    assert read.main([str(root), "--recursive", "--max-depth", "2"]) == 0

    output = capsys.readouterr().out
    assert f"# Directory: {root.resolve()}" in output
    assert "`.git/`" in output
    assert "`.git/config`" in output
    assert "`docs/`" in output
    assert "`docs/README.md`" in output


def test_read_requires_session_context_for_relative_hidden_file(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    repo = tmp_path / "repo"
    repo.mkdir()
    (repo / ".git").mkdir()
    config = repo / ".git" / "config"
    config.write_text("[remote \"origin\"]\n\turl = git@github.com:acme/widgets.git\n", encoding="utf-8")

    monkeypatch.chdir(tmp_path)
    monkeypatch.setenv("GOTTA_WORK_REPO", str(repo))

    assert read.main([".git/config"]) == 2
    err = capsys.readouterr().err
    assert "requires an active or explicit session root" in err


def test_read_prefers_session_relative_target_over_shell_cwd(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    cwd_root = tmp_path / "cwd"
    session_root = tmp_path / "session"
    cwd_root.mkdir()
    session_root.mkdir()
    (cwd_root / "BRIEF.md").write_text("cwd copy\n", encoding="utf-8")
    (session_root / "BRIEF.md").write_text("session copy\n", encoding="utf-8")

    monkeypatch.chdir(cwd_root)
    monkeypatch.setenv("GOTTA_SESSION_DIR", str(session_root))

    assert read.main(["BRIEF.md"]) == 0
    output = capsys.readouterr().out
    assert "session copy" in output
    assert "cwd copy" not in output


def test_read_can_follow_stored_content_digest(monkeypatch, tmp_path: Path, capsys) -> None:
    session_root = tmp_path / "session"
    dirs = content.ResolvedDirs(session_dir=session_root, content_dir=session_root / "content")
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    result = content.materialize_bytes(
        b"# Stored body\n\nline two\n",
        dirs=dirs,
        preferred_name="stored.md",
        metadata={"plugin": "read", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert read.main([result.digest]) == 0
    output = capsys.readouterr().out
    assert "# Stored body" in output


def test_read_can_follow_explicit_content_locator(monkeypatch, tmp_path: Path, capsys) -> None:
    session_root = tmp_path / "session"
    dirs = content.ResolvedDirs(session_dir=session_root, content_dir=session_root / "content")
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    result = content.materialize_bytes(
        b"# Stored body\n\nline two\n",
        dirs=dirs,
        preferred_name="stored.md",
        metadata={"plugin": "read", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert read.main([f"content:{result.digest}"]) == 0
    output = capsys.readouterr().out
    assert "# Stored body" in output


def test_read_can_follow_unique_session_artifact_name(monkeypatch, tmp_path: Path, capsys) -> None:
    session_root = tmp_path / "session"
    dirs = content.ResolvedDirs(session_dir=session_root, content_dir=session_root / "content")
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert read.main(["slack-search-abc.md"]) == 0
    output = capsys.readouterr().out
    assert "# Search Artifact" in output


def test_read_can_follow_explicit_artifact_locator(monkeypatch, tmp_path: Path, capsys) -> None:
    session_root = tmp_path / "session"
    dirs = content.ResolvedDirs(session_dir=session_root, content_dir=session_root / "content")
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    result = content.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert read.main([content.artifact_locator("slack-search-abc.md", result.digest)]) == 0
    output = capsys.readouterr().out
    assert "# Search Artifact" in output


def test_read_can_follow_artifact_locator_from_session_root_state_env(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content.ResolvedDirs(session_dir=session_root, content_dir=session_root / "content")
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content.write_state_env(dirs)
    result = content.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.delenv("GOTTA_SESSION_DIR", raising=False)
    monkeypatch.delenv("GOTTA_SESSION_CONTENT_DIR", raising=False)
    monkeypatch.chdir(dirs.session_dir)

    assert read.main([content.artifact_locator("slack-search-abc.md", result.digest)]) == 0
    output = capsys.readouterr().out
    assert "# Search Artifact" in output


def test_read_can_follow_artifact_locator_from_peer_root_state_env(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    dirs = content.ResolvedDirs(session_dir=session_root, content_dir=session_root / "content")
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    content.write_state_env(dirs)
    result = content.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )
    peer_root = session_root / "peers" / "claude"
    peer_state = peer_root / "state"
    peer_state.mkdir(parents=True, exist_ok=True)
    (peer_state / "env").write_text(
        "\n".join(
            [
                f"export GOTTA_SESSION_DIR='{peer_root}'",
                f"export GOTTA_SESSION_CONTENT_DIR='{dirs.content_dir}'",
            ]
        )
        + "\n",
        encoding="utf-8",
    )

    monkeypatch.delenv("GOTTA_SESSION_DIR", raising=False)
    monkeypatch.delenv("GOTTA_SESSION_CONTENT_DIR", raising=False)
    monkeypatch.chdir(peer_root)

    assert read.main([content.artifact_locator("slack-search-abc.md", result.digest)]) == 0
    output = capsys.readouterr().out
    assert "# Search Artifact" in output


def test_read_missing_peer_local_surface_reports_missing_local_path(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    peer_root = tmp_path / "session-claude"
    session_root.mkdir()
    peer_root.mkdir()
    (session_root / "peers").mkdir()
    (session_root / "peers" / "claude").symlink_to(
        os.path.relpath(peer_root, start=(session_root / "peers")),
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(session_root))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(session_root / "content"))

    assert read.main(["peers/claude/WANT.md"]) == 2
    err = capsys.readouterr().err
    assert "local target 'peers/claude/WANT.md' does not exist" in err
    assert "unsupported target" not in err


def test_read_missing_peer_goal_surface_reports_missing_local_path(
    monkeypatch, tmp_path: Path, capsys
) -> None:
    session_root = tmp_path / "session"
    peer_root = tmp_path / "session-claude"
    session_root.mkdir()
    peer_root.mkdir()
    (session_root / "peers").mkdir()
    (session_root / "peers" / "claude").symlink_to(
        os.path.relpath(peer_root, start=(session_root / "peers")),
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(session_root))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(session_root / "content"))

    assert read.main(["peers/claude/GOAL.md"]) == 2
    err = capsys.readouterr().err
    assert "local target 'peers/claude/GOAL.md' does not exist" in err
    assert "unsupported target" not in err


def test_read_seeded_peer_local_charter_surfaces(monkeypatch, tmp_path: Path, capsys) -> None:
    root = tmp_path / "work-root"

    assert session.main(["init", "--session", str(root)]) == 0
    capsys.readouterr()
    (root / "WANT.md").write_text("# Want\n\nReal intent.\n", encoding="utf-8")
    (root / "GOAL.md").write_text("# Goal\n\nReal goal.\n", encoding="utf-8")

    assert peer.main(["with", "Claude", "--session", str(root)]) == 0
    capsys.readouterr()

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(root))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(root / "content"))

    assert read.main(["peers/claude/WANT.md"]) == 0
    want_output = capsys.readouterr().out
    assert "# Peer Want Placeholder" in want_output

    assert read.main(["peers/claude/GOAL.md"]) == 0
    goal_output = capsys.readouterr().out
    assert "# Seed Peer Goal Placeholder" in goal_output


def test_read_does_not_materialize_local_artifact_rereads(monkeypatch, tmp_path: Path) -> None:
    session_root = tmp_path / "session"
    dirs = content.ResolvedDirs(session_dir=session_root, content_dir=session_root / "content")
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    result = content.materialize_bytes(
        b"# Search Artifact\n\nbody\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo", "canonical_locator": "demo"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    assert not target.should_materialize(
        [content.artifact_locator("slack-search-abc.md", result.digest), "--head", "2"]
    )


def test_read_ambiguous_session_artifact_name_suggests_explicit_artifact_locators(
    monkeypatch, tmp_path: Path
) -> None:
    session_root = tmp_path / "session"
    dirs = content.ResolvedDirs(session_dir=session_root, content_dir=session_root / "content")
    dirs.session_dir.mkdir(parents=True, exist_ok=True)
    dirs.content_dir.mkdir(parents=True, exist_ok=True)
    first = content.materialize_bytes(
        b"# Search Artifact A\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo-a", "canonical_locator": "demo-a"},
    )
    second = content.materialize_bytes(
        b"# Search Artifact B\n",
        dirs=dirs,
        preferred_name="slack-search-abc.md",
        metadata={"plugin": "slack", "locator": "demo-b", "canonical_locator": "demo-b"},
    )

    monkeypatch.setenv("GOTTA_SESSION_DIR", str(dirs.session_dir))
    monkeypatch.setenv("GOTTA_SESSION_CONTENT_DIR", str(dirs.content_dir))

    with pytest.raises(SystemExit) as excinfo:
        read.main(["slack-search-abc.md"])
    message = str(excinfo.value)
    assert "artifact:slack-search-abc.md@" in message
    assert content.artifact_locator("slack-search-abc.md", first.digest) in message
    assert content.artifact_locator("slack-search-abc.md", second.digest) in message


def test_read_supports_bounded_local_views(tmp_path: Path, capsys) -> None:
    target = tmp_path / "notes.md"
    target.write_text(
        "# Intro\n\nline 1\nline 2\n## Details\n\na\nb\nc\n",
        encoding="utf-8",
    )

    assert read.main([str(target), "--section", "Details", "--head", "2"]) == 0
    output = capsys.readouterr().out
    assert "## Details" in output
    assert "a" in output
    assert "b" not in output


def test_read_supports_bounded_provider_views(monkeypatch, capsys) -> None:
    monkeypatch.setattr(
        read,
        "resolve_read_target",
        lambda argv: target.ReadTarget(
            request=target.parse_args(argv),
            kind="routed",
            path=None,
            routed_plugin="github",
            routed_argv=["fake"],
            canonical_locator="https://github.com/acme/widgets",
            preferred_name="widgets.md",
            should_materialize=True,
        ),
    )

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        print("# Title")
        print("")
        print("line 1")
        print("line 2")
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["https://github.com/acme/widgets", "--head", "3"]) == 0
    output = capsys.readouterr().out
    assert "# Title" in output
    assert "line 1" in output
    assert "line 2" not in output


def test_read_passes_provider_flags_through_for_routed_targets(monkeypatch) -> None:
    resolved = target.resolve_read_target(
        ["--limit", "10", "https://github.com/acme/widgets/commits/main"]
    )

    assert resolved.kind == "routed"
    assert resolved.routed_plugin == "github"
    assert resolved.routed_argv == [
        "--limit",
        "10",
        "https://github.com/acme/widgets/commits/main",
    ]


def test_read_passes_provider_flags_through_after_routed_target(monkeypatch) -> None:
    resolved = target.resolve_read_target(
        ["https://github.com/acme/widgets/commits/main", "--limit", "10"]
    )

    assert resolved.kind == "routed"
    assert resolved.routed_plugin == "github"
    assert resolved.routed_argv == [
        "https://github.com/acme/widgets/commits/main",
        "--limit",
        "10",
    ]


def test_read_uses_provider_normalized_fragmentless_url_for_routed_targets() -> None:
    resolved = target.resolve_read_target(
        ["https://docs.google.com/spreadsheets/d/sheet-123/edit#gid=0"]
    )

    assert resolved.kind == "routed"
    assert resolved.routed_plugin == "gsheets"
    assert resolved.routed_argv == [
        "get",
        "https://docs.google.com/spreadsheets/d/sheet-123/edit",
    ]


def test_read_preserves_github_fragment_hint_while_canonicalizing_identity() -> None:
    resolved = target.resolve_read_target(["https://github.com/acme/widgets#readme"])

    assert resolved.kind == "routed"
    assert resolved.routed_plugin == "github"
    assert resolved.routed_argv == ["https://github.com/acme/widgets#readme"]
    assert resolved.canonical_locator == "https://github.com/acme/widgets"


def test_read_delegates_github_commit_history_limit_after_target(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["https://github.com/acme/widgets/commits/main", "--limit", "10"]) == 0
    assert seen == [
        (
            "github",
            ["https://github.com/acme/widgets/commits/main", "--limit", "10"],
        )
    ]


def test_read_can_follow_canonical_jira_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["jira:PROJ-3960"]) == 0
    assert seen == [("jira", ["get", "PROJ-3960"])]


def test_read_can_follow_canonical_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["jira:search Architecture"]) == 0
    assert seen == [("jira", ["search", "Architecture"])]


def test_read_can_follow_canonical_granola_document_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["granola:11111111-1111-1111-1111-111111111111"]) == 0
    assert seen == [("granola", ["get", "11111111-1111-1111-1111-111111111111"])]


def test_read_can_follow_canonical_granola_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["granola:search --limit 5 latency"]) == 0
    assert seen == [("granola", ["search", "--limit", "5", "latency"])]


def test_read_can_follow_canonical_granola_transcript_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["granola:search-transcript --all latency"]) == 0
    assert seen == [("granola", ["search-transcript", "--all", "latency"])]


def test_read_can_follow_canonical_granola_transcript_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["granola:transcript 11111111-1111-1111-1111-111111111111"]) == 0
    assert seen == [("granola", ["transcript", "11111111-1111-1111-1111-111111111111"])]


def test_read_can_follow_canonical_search_locator_with_flags(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert (
        read.main(["slack:search --workspace example-workspace --source archive --limit 10 --output markdown ABC"])
        == 0
    )
    assert seen == [
        (
            "slack",
            ["search", "--workspace", "example-workspace", "--source", "archive", "--limit", "10", "--output", "markdown", "ABC"],
        )
    ]


def test_read_can_follow_canonical_search_locator_with_equals_flags_unquoted(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["jira:search", "--limit=10", "Architecture"]) == 0
    assert seen == [("jira", ["search", "--limit", "10", "Architecture"])]


def test_read_can_follow_canonical_search_locator_with_flags_after_query(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["slack:search ABC reboot --workspace example-workspace --limit 10"]) == 0
    assert seen == [
        (
            "slack",
            ["search", "--workspace", "example-workspace", "--limit", "10", "ABC reboot"],
        )
    ]


def test_read_can_follow_canonical_github_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["github:search --type pr --repo acme/widgets ABC"]) == 0
    assert seen == [("github", ["search", "--type", "pr", "--repo", "acme/widgets", "ABC"])]


def test_read_can_follow_canonical_global_github_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["github:search --global ABC"]) == 0
    assert seen == [("github", ["search", "--global", "ABC"])]


def test_read_can_follow_slack_workspace_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["slack:example-workspace"]) == 0
    assert seen == [("slack", ["status", "--workspace", "example-workspace", "--output", "summary"])]


def test_read_joins_unquoted_multiword_search_locator(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["jira:search", "ABC", "reboot"]) == 0
    assert seen == [("jira", ["search", "ABC reboot"])]


def test_read_accepts_option_like_locator_tails_after_end_of_options(monkeypatch) -> None:
    seen: list[tuple[str, list[str]]] = []

    def fake_delegate(plugin: str, argv: list[str]) -> int:
        seen.append((plugin, argv))
        return 0

    monkeypatch.setattr(read, "delegate", fake_delegate)

    assert read.main(["--", "jira:search", "ABC", "--head", "180"]) == 0
    assert seen == [("jira", ["search", "ABC --head 180"])]



"""Shared locator-routing helpers for provider-owned read delegation."""

from __future__ import annotations

import shlex
import urllib.parse


def split_locator_tail(tail: str) -> list[str]:
    try:
        parts = shlex.split(tail)
    except ValueError:
        parts = [tail]
    return [part for part in parts if part]


def strip_http_url_fragment(target: str) -> str:
    parsed = urllib.parse.urlparse(target)
    if parsed.scheme not in {"http", "https"} or not parsed.fragment:
        return target
    return urllib.parse.urlunparse(parsed._replace(fragment=""))


def query_route(
    subcommand: str,
    tail: str,
    *,
    valued_flags: tuple[str, ...] = (),
    boolean_flags: tuple[str, ...] = (),
) -> list[str] | None:
    parts = split_locator_tail(tail)
    if not parts:
        return None
    argv = [subcommand]
    query_parts: list[str] = []
    index = 0
    parse_flags = True
    while index < len(parts):
        token = parts[index]
        if parse_flags and token == "--":
            parse_flags = False
            index += 1
            continue
        if parse_flags and token in boolean_flags:
            argv.append(token)
            index += 1
            continue
        if parse_flags and token in valued_flags:
            if index + 1 >= len(parts):
                return None
            argv.extend([token, parts[index + 1]])
            index += 2
            continue
        if parse_flags:
            matched_flag = next(
                (flag for flag in valued_flags if token.startswith(f"{flag}=")),
                "",
            )
            if matched_flag:
                argv.extend([matched_flag, token.split("=", 1)[1]])
                index += 1
                continue
        query_parts.append(token)
        index += 1
    query = " ".join(part for part in query_parts if part).strip()
    if not query:
        return None
    return [*argv, query]



"""Compatibility shims for supported Python versions."""

from __future__ import annotations

from datetime import datetime, timezone

try:
    from datetime import UTC
except ImportError:  # pragma: no cover
    UTC = timezone.utc

try:
    import tomllib
except ModuleNotFoundError:  # pragma: no cover
    import tomli as tomllib

__all__ = ["UTC", "datetime", "tomllib"]



#!/usr/bin/env python3
"""Standalone Confluence page-body toolbelt."""

from __future__ import annotations

import argparse
import difflib
import html
import json
import os
import re
import shlex
import shutil
import signal
import subprocess
import sys
import urllib.parse
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from gotta.config import set_provider_env_values
from gotta.helptext import is_long_help_request, print_long_help
from gotta.routing import query_route, strip_http_url_fragment
from gotta.source import derive_source_metadata_from_payload, render_source_metadata_lines
from gotta.providers import atlassian as atl


DEFAULT_FIND_CONTEXT_CHARS = 120
DEFAULT_CONFLUENCE_BASE_URL = ""
CONFLUENCE_BASE_URL_ENV = "GOTTA_CONFLUENCE_BASE_URL"
PAGE_ARG_HELP = (
    "Confluence page URL, space landing/overview URL, or numeric page ID "
    f"against {CONFLUENCE_BASE_URL_ENV} when a Confluence base URL is persisted"
)
DISALLOWED_MCP_PASSTHROUGH_FLAGS = atl.DISALLOWED_MCP_PASSTHROUGH_FLAGS
ToolError = atl.AtlassianError


@dataclass
class PageRef:
    page_id: str | None = None
    base_url: str = ""
    space_key: str = ""


@dataclass
class ContentRef:
    requested_id: str | None = None
    page_id: str | None = None
    comment_id: str | None = None
    base_url: str = ""
    space_key: str = ""


@dataclass
class Session:
    token: str
    cloud_id: str
    base_url: str


def die(message: str, code: int = 2) -> int:
    print(message, file=sys.stderr)
    return code


site_root = atl.site_root
read_text = atl.read_text
load_atlassian_config_env = atl.load_atlassian_config_env
load_oauth_runtime_config = atl.load_oauth_runtime_config
is_interactive = atl.is_interactive
api_json = atl.api_json
token_preflight_status = atl.token_preflight_status
load_cloud_id = atl.load_cloud_id
atlassian_status_payload = atl.atlassian_status_payload


def default_base_url() -> str:
    config_env = load_atlassian_config_env()
    return atl.env_or_config(
        config_env,
        CONFLUENCE_BASE_URL_ENV,
        default=DEFAULT_CONFLUENCE_BASE_URL,
    ).strip()


def _slug(value: str, *, fallback: str) -> str:
    value = value.strip().lower()
    value = re.sub(r"[^a-z0-9._-]+", "-", value)
    return value.strip("-") or fallback


def _output_extension(output: str) -> str:
    return {
        "markdown": "md",
        "json": "json",
        "body": "body.html",
        "meta": "json",
        "summary": "summary",
        "text": "txt",
    }.get(output, "md")


def _parse_cli(argv: list[str]) -> argparse.Namespace:
    return build_parser().parse_args(argv)


def canonical_locator(argv: list[str]) -> str:
    args = _parse_cli(argv)
    if args.command == "get":
        content_ref = parse_content_ref(args.page)
        if content_ref.comment_id:
            return f"confluence:{content_ref.comment_id}"
        if content_ref.requested_id:
            return f"confluence:{content_ref.requested_id}"
        return f"confluence:{args.page}"
    if args.command == "search":
        return f"confluence:search {args.query}"
    if args.command == "cql":
        return f"confluence:cql {args.query}"
    if args.command == "status":
        return "confluence:status"
    return f"confluence:{shlex.join(argv)}"


def preferred_name(argv: list[str], options: object) -> str:
    if getattr(options, "save_as", ""):
        return str(getattr(options, "save_as"))
    args = _parse_cli(argv)
    if args.command == "get":
        content_ref = parse_content_ref(args.page)
        content_id = content_ref.comment_id or content_ref.requested_id or ""
        if content_id:
            return f"{content_id}.{_output_extension(args.output)}"
        parsed = urllib.parse.urlparse(args.page)
        page_name = Path(parsed.path.rstrip("/")).name if parsed.scheme else args.page
        return f"{_slug(page_name, fallback='confluence')}.{_output_extension(args.output)}"
    if args.command == "search":
        return (
            f"confluence-search-{_slug(args.query, fallback='confluence')}"
            f".{_output_extension(args.output)}"
        )
    if args.command == "cql":
        return (
            f"confluence-cql-{_slug(args.query, fallback='confluence')}"
            f".{_output_extension('markdown')}"
        )
    if args.command == "status":
        return f"confluence.{_output_extension(args.output)}"
    return "confluence.txt"


def route_target(target: str) -> list[str] | None:
    if target.startswith("https://") and ".atlassian.net/wiki/" in target:
        if any(char.isspace() for char in target):
            return None
        return ["get", strip_http_url_fragment(target)]
    if target.startswith("confluence:search "):
        return query_route(
            "search",
            target.removeprefix("confluence:search "),
            valued_flags=("--space", "--type", "--limit", "--next", "--output", "--base-url"),
            boolean_flags=("--title-only",),
        )
    if target.startswith("confluence:cql "):
        return query_route(
            "cql",
            target.removeprefix("confluence:cql "),
            valued_flags=("--limit", "--next", "--base-url"),
        )
    if target.startswith("confluence:"):
        return ["get", target.removeprefix("confluence:")]
    return None


def persist_selected_base_urls(base_url: str) -> None:
    confluence_url = base_url.strip().rstrip("/")
    site_url = site_root(confluence_url)
    if not site_url:
        return
    set_provider_env_values(
        "atlassian",
        {
            "GOTTA_JIRA_BASE_URL": site_url,
            CONFLUENCE_BASE_URL_ENV: confluence_url or f"{site_url}/wiki",
        },
    )


def run_oauth_bootstrap(*, base_url: str = "") -> dict[str, Any]:
    return atl.run_oauth_bootstrap(
        base_url=base_url,
        base_url_env=CONFLUENCE_BASE_URL_ENV,
    )


def parse_page_ref(raw: str) -> PageRef:
    if raw.startswith("confluence:"):
        raw = raw.removeprefix("confluence:")
    page_id: str | None = atl.extract_confluence_page_id(raw)
    space_key = ""
    base_url = default_base_url()
    base_match = re.match(r"^(https://[^/]+\.atlassian\.net)(?:/wiki/.*)?$", raw)
    if base_match:
        base_url = f"{base_match.group(1)}/wiki"
    if not page_id:
        overview_match = re.match(
            r"^(https://[^/]+\.atlassian\.net)/wiki/spaces/([^/]+)(?:/overview)?/?$",
            raw,
        )
        if overview_match:
            base_url = overview_match.group(1)
            space_key = urllib.parse.unquote(overview_match.group(2))
    if page_id is None and not space_key:
        raise ToolError(f"could not parse Confluence page ID from input: {raw}")
    return PageRef(page_id=page_id, base_url=base_url, space_key=space_key)


def _extract_focused_comment_id(raw: str) -> str:
    parsed = urllib.parse.urlparse(raw.strip())
    params = urllib.parse.parse_qs(parsed.query)
    return str(params.get("focusedCommentId", [""])[0] or "").strip()


def parse_content_ref(raw: str) -> ContentRef:
    candidate = raw.strip()
    if candidate.startswith("confluence:"):
        candidate = candidate.removeprefix("confluence:")
    base_url = default_base_url()
    base_match = re.match(r"^(https://[^/]+\.atlassian\.net)(?:/wiki/.*)?$", candidate)
    if base_match:
        base_url = f"{base_match.group(1)}/wiki"
    focused_comment_id = _extract_focused_comment_id(candidate)
    if focused_comment_id.isdigit():
        return ContentRef(
            requested_id=focused_comment_id,
            page_id=atl.extract_confluence_page_id(candidate),
            comment_id=focused_comment_id,
            base_url=base_url,
        )
    page_ref = parse_page_ref(candidate)
    requested_id = page_ref.page_id
    return ContentRef(
        requested_id=requested_id,
        page_id=page_ref.page_id,
        base_url=page_ref.base_url,
        space_key=page_ref.space_key,
    )


def load_token(base_url: str = "") -> str:
    return atl.load_token(
        base_url=base_url,
        auth_command="confluence",
        base_url_env=CONFLUENCE_BASE_URL_ENV,
    )


def should_reauth_after_api_error(token: str) -> bool:
    return token_preflight_status(token) == "invalid"


def discover_cloud_id(token: str, base_url: str) -> str:
    return atl.discover_cloud_id(
        token,
        base_url,
        base_url_env=CONFLUENCE_BASE_URL_ENV,
    )


def load_session(page_ref: PageRef, *, allow_reauth: bool = True) -> Session:
    token = load_token(page_ref.base_url)
    try:
        if page_ref.base_url:
            cloud_id = discover_cloud_id(token, page_ref.base_url)
            resolved_base_url = site_root(page_ref.base_url)
        else:
            cloud_id, resolved_base_url = atl.resolve_accessible_resource(
                token,
                "",
                base_url_env=CONFLUENCE_BASE_URL_ENV,
                cloud_id=load_cloud_id(),
            )
        return Session(token=token, cloud_id=cloud_id, base_url=resolved_base_url)
    except ToolError:
        if allow_reauth and is_interactive() and token_preflight_status(token) == "invalid":
            run_oauth_bootstrap(base_url=page_ref.base_url)
            return load_session(page_ref, allow_reauth=False)
        raise


def page_api_url(session: Session, page_id: str, *, body_format: str = "storage") -> str:
    params = urllib.parse.urlencode({"body-format": body_format})
    return (
        f"https://api.atlassian.com/ex/confluence/{session.cloud_id}/wiki/api/v2/"
        f"pages/{page_id}?{params}"
    )


def pages_api_url(
    session: Session,
    *,
    title: str = "",
    space_id: str = "",
    status: str = "",
    limit: int = 25,
    cursor: str | None = None,
) -> str:
    params: dict[str, Any] = {"limit": limit}
    if title:
        params["title"] = title
    if space_id:
        params["space-id"] = space_id
    if status:
        params["status"] = status
    if cursor:
        params["cursor"] = cursor
    query = urllib.parse.urlencode(params, doseq=True)
    return (
        f"https://api.atlassian.com/ex/confluence/{session.cloud_id}/wiki/api/v2/"
        f"pages?{query}"
    )


def pages_collection_url(session: Session) -> str:
    return f"https://api.atlassian.com/ex/confluence/{session.cloud_id}/wiki/api/v2/pages"


def comment_api_url(
    session: Session,
    comment_id: str,
    *,
    comment_kind: str,
    body_format: str = "storage",
) -> str:
    params = urllib.parse.urlencode({"body-format": body_format})
    return (
        f"https://api.atlassian.com/ex/confluence/{session.cloud_id}/wiki/api/v2/"
        f"{comment_kind}/{comment_id}?{params}"
    )


def search_api_url(
    session: Session, cql: str, *, limit: int, cursor: str | None
) -> str:
    params: dict[str, Any] = {"cql": cql, "limit": limit}
    if cursor:
        params["cursor"] = cursor
    query = urllib.parse.urlencode(params, doseq=True)
    return (
        f"https://api.atlassian.com/ex/confluence/{session.cloud_id}/wiki/rest/api/"
        f"search?{query}"
    )


def cql_string_literal(value: str) -> str:
    escaped = value.replace("\\", "\\\\").replace('"', '\\"')
    return f'"{escaped}"'


def absolutize_confluence_url(url: str, *, base_url: str) -> str:
    if not url:
        return ""
    if re.match(r"^https?://", url):
        return url
    if base_url:
        return urllib.parse.urljoin(base_url.rstrip("/") + "/", url.lstrip("/"))
    return url


def extract_cursor(link: str) -> str:
    parsed = urllib.parse.urlparse(link)
    params = urllib.parse.parse_qs(parsed.query)
    return str(params.get("cursor", [""])[0] or "")


def build_text_search_cql(args: argparse.Namespace) -> str:
    clauses: list[str] = []
    field = "title" if args.title_only else "text"
    clauses.append(f"{field} ~ {cql_string_literal(args.query)}")
    if args.space:
        clauses.append(f"space = {cql_string_literal(args.space)}")
    if args.type and args.type != "all":
        clauses.append(f"type = {args.type}")
    return " AND ".join(clauses)


def normalize_search_result(
    result: dict[str, Any], *, base_url: str
) -> dict[str, Any]:
    content = result.get("content")
    if not isinstance(content, dict):
        content = {}
    space = content.get("space")
    if not isinstance(space, dict):
        space = {}
    container = result.get("container")
    if not isinstance(container, dict):
        container = {}
    normalized = {
        "title": result.get("title") or content.get("title"),
        "id": content.get("id") or result.get("id"),
        "type": content.get("type") or result.get("entityType"),
        "url": absolutize_confluence_url(str(result.get("url") or ""), base_url=base_url),
    }
    excerpt = result.get("excerpt")
    if excerpt:
        normalized["excerptHtml"] = excerpt
    if space.get("key"):
        normalized["spaceKey"] = space.get("key")
    if space.get("name"):
        normalized["spaceName"] = space.get("name")
    if container.get("title"):
        normalized["containerTitle"] = container.get("title")
    if result.get("lastModified"):
        normalized["lastModified"] = result.get("lastModified")
    if result.get("friendlyLastModified"):
        normalized["friendlyLastModified"] = result.get("friendlyLastModified")
    return normalized


def render_excerpt_text(excerpt_html: str) -> str:
    text = re.sub(r"<[^>]+>", " ", excerpt_html)
    text = html.unescape(text)
    text = re.sub(r"\s+", " ", text).strip()
    return text


def render_search_markdown(payload: dict[str, Any]) -> str:
    results = payload.get("results")
    if not isinstance(results, list) or not results:
        return (
            f"### Confluence Search: {payload.get('cql') or ''}\n\n"
            f"- _Type_: `{payload.get('type') or 'page'}`\n"
            f"- _Matches_: 0\n\n"
            "No Confluence content matched.\n"
        )
    lines: list[str] = [
        f"### Confluence Search: {payload.get('cql') or ''}",
        "",
        "- _Surface_: `confluence`",
        f"- _Type_: `{payload.get('type') or 'page'}`",
        f"- _Matches_: {payload.get('size') or len(results)}",
    ]
    lines.extend(render_source_metadata_lines(derive_source_metadata_from_payload(payload)))
    if payload.get("next"):
        lines.append(f"- _Next_: `{payload['next']}`")
    if payload.get("previous"):
        lines.append(f"- _Previous_: `{payload['previous']}`")
    lines.append("")
    for item in results:
        if not isinstance(item, dict):
            continue
        title = str(item.get("title") or "(untitled)")
        url = str(item.get("url") or "")
        content_id = str(item.get("id") or "")
        content_type = str(item.get("type") or "")
        space_key = str(item.get("spaceKey") or "")
        space_name = str(item.get("spaceName") or "")
        modified = str(item.get("lastModified") or item.get("friendlyLastModified") or "")
        line = f"- [{title}]({url})"
        details: list[str] = []
        if content_id:
            details.append(f"locator `confluence:{content_id}`")
        if content_type:
            details.append(f"type `{content_type}`")
        if space_key:
            details.append(f"space `{space_key}`")
        elif space_name:
            details.append(f"space `{space_name}`")
        if modified:
            details.append(f"modified `{modified}`")
        if details:
            line += " - " + ", ".join(details)
        lines.append(line)
        excerpt = str(item.get("excerptHtml") or "").strip()
        if excerpt:
            lines.append(f"  - {render_excerpt_text(excerpt)}")
    return "\n".join(lines) + "\n"


def search_confluence(
    session: Session,
    *,
    cql: str,
    limit: int,
    cursor: str | None,
    allow_reauth: bool = True,
) -> dict[str, Any]:
    try:
        payload = api_json(
            "GET",
            search_api_url(session, cql, limit=limit, cursor=cursor),
            session.token,
        )
        if not isinstance(payload, dict):
            raise ToolError("unexpected search response")
        links = payload.get("_links")
        if not isinstance(links, dict):
            links = {}
        results = payload.get("results")
        if not isinstance(results, list):
            results = []
        base_url = str(links.get("base") or session.base_url or "").strip()
        next_link = str(links.get("next") or "")
        prev_link = str(links.get("prev") or "")
        type_match = re.search(r"type = ([a-z]+)", cql or "")
        response = {
            "baseUrl": base_url,
            "cql": cql,
            "limit": limit,
            "size": len(results),
            "type": type_match.group(1) if type_match else "",
            "results": [
                normalize_search_result(item, base_url=base_url)
                for item in results
                if isinstance(item, dict)
            ],
        }
        if cursor:
            response["requestedNext"] = cursor
        if next_link:
            response["next"] = extract_cursor(next_link)
        if prev_link:
            response["previous"] = extract_cursor(prev_link)
        return response
    except ToolError:
        if (
            allow_reauth
            and is_interactive()
            and should_reauth_after_api_error(session.token)
        ):
            run_oauth_bootstrap(base_url=session.base_url)
            refreshed_session = load_session(
                PageRef(base_url=session.base_url),
                allow_reauth=False,
            )
            return search_confluence(
                refreshed_session,
                cql=cql,
                limit=limit,
                cursor=cursor,
                allow_reauth=False,
            )
        raise


def resolve_page_id(session: Session, page_ref: PageRef) -> str:
    if page_ref.page_id:
        return page_ref.page_id
    if not page_ref.space_key:
        raise ToolError("missing page id and space key")
    params = urllib.parse.urlencode({"keys": [page_ref.space_key]}, doseq=True)
    url = (
        f"https://api.atlassian.com/ex/confluence/{session.cloud_id}/wiki/api/v2/"
        f"spaces?{params}"
    )
    payload = api_json("GET", url, session.token)
    if not isinstance(payload, dict):
        raise ToolError("unexpected space lookup response")
    results = payload.get("results")
    if not isinstance(results, list) or len(results) != 1 or not isinstance(results[0], dict):
        raise ToolError(
            f"could not resolve unique Confluence homepage for space {page_ref.space_key}"
        )
    homepage_id = str(results[0].get("homepageId") or "").strip()
    if not homepage_id:
        raise ToolError(
            f"space {page_ref.space_key} does not expose a homepageId via the Confluence API"
        )
    return homepage_id


def fetch_page(
    page_ref: PageRef, *, allow_reauth: bool = True
) -> tuple[Session, dict[str, Any]]:
    session: Session | None = None
    try:
        session = load_session(page_ref, allow_reauth=allow_reauth)
        page_id = resolve_page_id(session, page_ref)
        page = _fetch_page_payload(session, page_id)
        return session, page
    except ToolError:
        if (
            allow_reauth
            and is_interactive()
            and session is not None
            and should_reauth_after_api_error(session.token)
        ):
            run_oauth_bootstrap(base_url=page_ref.base_url)
            return fetch_page(page_ref, allow_reauth=False)
        raise


def _fetch_page_payload(session: Session, page_id: str) -> dict[str, Any]:
    page = api_json("GET", page_api_url(session, page_id), session.token)
    if not isinstance(page, dict):
        raise ToolError("unexpected page response")
    return page


def page_web_url(session: Session, page_id: str) -> str:
    if not session.base_url or not page_id:
        return ""
    return f"{session.base_url.rstrip('/')}/wiki/pages/viewpage.action?pageId={page_id}"


def normalize_page_summary(page: dict[str, Any], session: Session) -> dict[str, Any]:
    normalized = {
        "id": str(page.get("id") or ""),
        "title": str(page.get("title") or ""),
        "status": str(page.get("status") or ""),
        "spaceId": str(page.get("spaceId") or ""),
        "parentId": str(page.get("parentId") or ""),
        "url": page_web_url(session, str(page.get("id") or "")),
    }
    version = page.get("version")
    if isinstance(version, dict):
        normalized["version"] = version
    created = str(page.get("createdAt") or "")
    if created:
        normalized["createdAt"] = created
    return normalized


def list_pages(
    session: Session,
    *,
    title: str = "",
    space_id: str = "",
    status: str = "current",
    limit: int = 25,
    cursor: str | None = None,
) -> tuple[list[dict[str, Any]], str]:
    payload = api_json(
        "GET",
        pages_api_url(
            session,
            title=title,
            space_id=space_id,
            status=status,
            limit=limit,
            cursor=cursor,
        ),
        session.token,
    )
    if not isinstance(payload, dict):
        raise ToolError("unexpected pages response")
    results = payload.get("results")
    if not isinstance(results, list):
        results = []
    links = payload.get("_links")
    if not isinstance(links, dict):
        links = {}
    next_link = str(links.get("next") or "")
    return [item for item in results if isinstance(item, dict)], extract_cursor(next_link)


def find_child_pages_by_title(
    session: Session,
    *,
    parent_page: dict[str, Any],
    title: str,
) -> list[dict[str, Any]]:
    parent_id = str(parent_page.get("id") or "")
    space_id = str(parent_page.get("spaceId") or "")
    if not parent_id or not space_id:
        raise ToolError("parent page is missing required id/spaceId metadata")
    matches: list[dict[str, Any]] = []
    cursor: str | None = None
    while True:
        pages, cursor = list_pages(
            session,
            title=title,
            space_id=space_id,
            status="current",
            limit=100,
            cursor=cursor,
        )
        for page in pages:
            if str(page.get("title") or "") != title:
                continue
            if str(page.get("parentId") or "") != parent_id:
                continue
            matches.append(page)
        if not cursor:
            break
    return matches


def summarize_storage_html(storage_html: str, *, preview_chars: int = 240) -> dict[str, Any]:
    normalized = re.sub(r"\s+", " ", storage_html).strip()
    preview = normalized[:preview_chars]
    if len(normalized) > preview_chars:
        preview += "..."
    return {
        "chars": len(storage_html),
        "preview": preview,
    }


def create_page(
    session: Session,
    *,
    parent_page: dict[str, Any],
    title: str,
    body_html: str,
    allow_reauth: bool = True,
) -> dict[str, Any]:
    parent_id = str(parent_page.get("id") or "")
    space_id = str(parent_page.get("spaceId") or "")
    if not parent_id or not space_id:
        raise ToolError("parent page is missing required id/spaceId metadata")
    payload = {
        "spaceId": space_id,
        "status": "current",
        "title": title,
        "parentId": parent_id,
        "body": {"representation": "storage", "value": body_html},
    }
    try:
        created = api_json(
            "POST",
            pages_collection_url(session),
            session.token,
            payload=payload,
        )
        if not isinstance(created, dict):
            raise ToolError("unexpected create response")
        page_id = str(created.get("id") or "")
        if not page_id:
            raise ToolError("create response did not include a page id")
        persisted = api_json("GET", page_api_url(session, page_id), session.token)
        if not isinstance(persisted, dict):
            raise ToolError("unexpected persisted page response after create")
        return persisted
    except ToolError:
        if (
            allow_reauth
            and is_interactive()
            and should_reauth_after_api_error(session.token)
        ):
            run_oauth_bootstrap(base_url=session.base_url)
            refreshed_session = load_session(
                PageRef(page_id=parent_id, base_url=session.base_url),
                allow_reauth=False,
            )
            return create_page(
                refreshed_session,
                parent_page=parent_page,
                title=title,
                body_html=body_html,
                allow_reauth=False,
            )
        raise


def _is_not_found_error(exc: ToolError) -> bool:
    return " failed with 404:" in str(exc)


def _fetch_comment_payload(
    session: Session,
    comment_id: str,
    *,
    page_id: str = "",
) -> dict[str, Any]:
    last_error: ToolError | None = None
    for comment_kind in ("footer-comments", "inline-comments"):
        try:
            comment = api_json(
                "GET",
                comment_api_url(session, comment_id, comment_kind=comment_kind),
                session.token,
            )
        except ToolError as exc:
            last_error = exc
            if _is_not_found_error(exc):
                continue
            raise
        if not isinstance(comment, dict):
            raise ToolError("unexpected comment response")
        comment["_comment_kind"] = comment_kind.removesuffix("s")
        if page_id and not str(comment.get("pageId") or "").strip():
            comment["pageId"] = page_id
        return comment
    if last_error is not None:
        raise last_error
    raise ToolError(f"could not resolve Confluence comment {comment_id}")


def fetch_read_target(
    content_ref: ContentRef, *, allow_reauth: bool = True
) -> tuple[Session, str, dict[str, Any]]:
    session: Session | None = None
    try:
        session = load_session(PageRef(base_url=content_ref.base_url), allow_reauth=allow_reauth)
        if content_ref.comment_id:
            comment = _fetch_comment_payload(
                session,
                content_ref.comment_id,
                page_id=str(content_ref.page_id or ""),
            )
            return session, "comment", comment
        page_ref = PageRef(
            page_id=content_ref.page_id,
            base_url=content_ref.base_url,
            space_key=content_ref.space_key,
        )
        try:
            page_id = resolve_page_id(session, page_ref)
            page = _fetch_page_payload(session, page_id)
            return session, "page", page
        except ToolError as exc:
            if not content_ref.requested_id or not _is_not_found_error(exc):
                raise
            comment = _fetch_comment_payload(session, content_ref.requested_id)
            return session, "comment", comment
    except ToolError:
        if (
            allow_reauth
            and is_interactive()
            and session is not None
            and should_reauth_after_api_error(session.token)
        ):
            run_oauth_bootstrap(base_url=content_ref.base_url)
            return fetch_read_target(content_ref, allow_reauth=False)
        raise


def storage_value(page: dict[str, Any]) -> str:
    return str(page["body"]["storage"]["value"])


def comment_storage_value(comment: dict[str, Any]) -> str:
    return str(comment["body"]["storage"]["value"])


def update_page(
    session: Session,
    page: dict[str, Any],
    new_body: str,
    *,
    message: str,
    allow_reauth: bool = True,
) -> dict[str, Any]:
    try:
        payload = {
            "id": str(page["id"]),
            "status": "current",
            "title": page["title"],
            "body": {"representation": "storage", "value": new_body},
            "version": {
                "number": int(page["version"]["number"]) + 1,
                "message": message,
                "minorEdit": True,
            },
        }
        url = (
            f"https://api.atlassian.com/ex/confluence/{session.cloud_id}/wiki/api/v2/"
            f"pages/{page['id']}"
        )
        updated = api_json("PUT", url, session.token, payload=payload)
        if not isinstance(updated, dict):
            raise ToolError("unexpected update response")
        persisted = api_json("GET", page_api_url(session, str(page["id"])), session.token)
        if not isinstance(persisted, dict):
            raise ToolError("unexpected persisted page response after update")
        expected_version = int(page["version"]["number"]) + 1
        persisted_version = int(persisted["version"]["number"])
        if persisted_version != expected_version:
            raise ToolError(
                "persisted page version did not match update request: "
                f"expected {expected_version}, got {persisted_version}"
            )
        persisted["_update_echo_version"] = updated.get("version")
        return persisted
    except ToolError:
        if (
            allow_reauth
            and is_interactive()
            and should_reauth_after_api_error(session.token)
        ):
            run_oauth_bootstrap(base_url=session.base_url)
            refreshed_session = load_session(
                PageRef(page_id=str(page["id"]), base_url=session.base_url),
                allow_reauth=False,
            )
            return update_page(
                refreshed_session,
                page,
                new_body,
                message=message,
                allow_reauth=False,
            )
        raise


def ensure_base_version(page: dict[str, Any], expected_version: int | None) -> None:
    if expected_version is None:
        return
    current_version = int(page["version"]["number"])
    if current_version != expected_version:
        raise ToolError(
            "page version changed before update: "
            f"expected {expected_version}, got {current_version}"
        )


def format_update_result(updated: dict[str, Any]) -> dict[str, Any]:
    result = {
        "id": updated["id"],
        "title": updated["title"],
        "version": updated["version"],
    }
    echoed = updated.get("_update_echo_version")
    if isinstance(echoed, dict):
        result["updateEchoVersion"] = echoed
    return result


def replace_exact(body: str, old: str, new: str, *, expected_count: int) -> str:
    count = body.count(old)
    if count != expected_count:
        raise ToolError(
            f"expected {expected_count} match(es) for exact fragment, found {count}"
        )
    return body.replace(old, new)


def replace_between(
    body: str,
    start_marker: str,
    end_marker: str,
    replacement: str,
    *,
    include_markers: bool,
) -> str:
    start = body.find(start_marker)
    if start == -1:
        raise ToolError("start marker not found")
    search_from = start + len(start_marker)
    end = body.find(end_marker, search_from)
    if end == -1:
        raise ToolError("end marker not found")
    if end <= start:
        raise ToolError("end marker occurs before start marker")
    if include_markers:
        return body[:start] + replacement + body[end + len(end_marker) :]
    return body[:search_from] + replacement + body[end:]


def apply_operation(body: str, operation: dict[str, Any], *, index: int) -> str:
    op_type = str(operation.get("type", "")).strip()
    label = str(operation.get("label", f"operation {index}")).strip() or f"operation {index}"
    try:
        if op_type == "replace":
            old = read_explicit_input(
                operation.get("from"),
                operation.get("from_file"),
                input_name=f"{label} from input",
            )
            new = read_explicit_input(
                operation.get("to"),
                operation.get("to_file"),
                input_name=f"{label} to input",
            )
            expected_count = int(operation.get("expected_count", 1))
            return replace_exact(body, old, new, expected_count=expected_count)
        if op_type == "replace-section":
            start = str(operation.get("start", ""))
            end = str(operation.get("end", ""))
            if not start or not end:
                raise ToolError("replace-section requires start and end markers")
            replacement = read_explicit_input(
                operation.get("replacement"),
                operation.get("replacement_file"),
                input_name=f"{label} replacement input",
            )
            include_markers = bool(operation.get("include_markers", False))
            return replace_between(
                body,
                start,
                end,
                replacement,
                include_markers=include_markers,
            )
    except ToolError as exc:
        raise ToolError(f"{label}: {exc}") from exc
    raise ToolError(f"{label}: unsupported operation type: {op_type or '<missing>'}")


def load_operations(path: str) -> list[dict[str, Any]]:
    try:
        payload = json.loads(read_text(Path(path)))
    except json.JSONDecodeError as exc:
        raise ToolError(f"invalid batch JSON in {path}: {exc}") from exc
    if not isinstance(payload, list) or not payload:
        raise ToolError("batch file must contain a non-empty JSON array of operations")
    operations: list[dict[str, Any]] = []
    for index, item in enumerate(payload, start=1):
        if not isinstance(item, dict):
            raise ToolError(f"operation {index}: expected JSON object")
        operations.append(item)
    return operations


def format_noop_result(page: dict[str, Any]) -> dict[str, Any]:
    return {
        "id": page["id"],
        "title": page["title"],
        "version": page["version"],
        "noop": True,
    }


def unified_diff(before: str, after: str, *, context: int = 2) -> str:
    lines = difflib.unified_diff(
        before.splitlines(),
        after.splitlines(),
        fromfile="before",
        tofile="after",
        n=context,
        lineterm="",
    )
    return "\n".join(lines)


def print_json(data: Any) -> None:
    json.dump(data, sys.stdout, indent=2, sort_keys=True)
    sys.stdout.write("\n")


def count_with_context(body: str, needle: str, *, context_chars: int) -> dict[str, Any]:
    hits = []
    start = 0
    while True:
        idx = body.find(needle, start)
        if idx == -1:
            break
        hits.append(
            {
                "offset": idx,
                "snippet": body[
                    max(0, idx - context_chars) : idx + len(needle) + context_chars
                ],
            }
        )
        start = idx + max(1, len(needle))
    return {"count": len(hits), "matches": hits}


def read_explicit_input(
    inline: str | None, file_path: str | None, *, input_name: str
) -> str:
    if inline is not None and file_path is not None:
        raise ToolError(f"use either inline {input_name} or file input, not both")
    if file_path is not None:
        return read_text(Path(file_path))
    if inline is not None:
        return inline
    raise ToolError(f"missing {input_name}")


def read_single_input(
    inline: str | None, file_path: str | None, *, input_name: str
) -> str:
    if inline is not None and file_path is not None:
        raise ToolError(f"use either inline {input_name} or file input, not both")
    if file_path is not None:
        return read_text(Path(file_path))
    if inline is not None:
        return inline
    if not sys.stdin.isatty():
        return sys.stdin.read()
    raise ToolError(f"missing {input_name}")


def render_markdown_to_storage(markdown: str) -> str:
    try:
        proc = subprocess.run(
            ["pandoc", "-f", "gfm", "-t", "html", "--wrap=none"],
            input=markdown,
            text=True,
            capture_output=True,
            check=True,
        )
    except FileNotFoundError as exc:
        raise ToolError("pandoc is required for markdown conversion") from exc
    except subprocess.CalledProcessError as exc:
        raise ToolError(f"pandoc failed: {exc.stderr.strip()}") from exc
    return proc.stdout


def _normalize_heading_title(text: str) -> str:
    return re.sub(r"\s+", " ", text.strip()).casefold()


def strip_matching_leading_h1(markdown: str, *, title: str) -> str:
    title_norm = _normalize_heading_title(title)
    if not title_norm:
        return markdown

    atx_match = re.match(
        r"^(?:[ \t]*\r?\n)*[ \t]{0,3}#(?!#)[ \t]+(?P<title>.*?)(?:[ \t]+#+)?[ \t]*(?:\r?\n|$)",
        markdown,
    )
    if atx_match and _normalize_heading_title(str(atx_match.group("title") or "")) == title_norm:
        return markdown[atx_match.end() :].lstrip("\r\n")

    setext_match = re.match(
        r"^(?:[ \t]*\r?\n)*(?P<title>[^\r\n]+)\r?\n[ \t]*=+[ \t]*(?:\r?\n|$)",
        markdown,
    )
    if setext_match and _normalize_heading_title(str(setext_match.group("title") or "")) == title_norm:
        return markdown[setext_match.end() :].lstrip("\r\n")

    return markdown


def _clean_markdown_projection(markdown: str) -> str:
    cleaned = re.sub(r"(?m)^wide760", "", markdown)
    cleaned = re.sub(
        r"(?m)^[0-9]+[ \t]+[0-9a-f]{8}-[0-9a-f-]{27}[ \t]+incomplete(?:\s+)*$",
        "",
        cleaned,
    )
    cleaned = re.sub(
        r"(?m)^[0-9]+(?:false|true)(?:default|list)(?:true|false)\s*$",
        "",
        cleaned,
    )
    cleaned = re.sub(
        r"(?m)^[0-9A-Za-z:/._-]*Untitled Diagram-[0-9]+\.drawio[0-9A-Za-z:/._-]*$",
        "",
        cleaned,
    )
    return cleaned


def _sanitize_storage_html_for_markdown(storage_html: str) -> str:
    cleaned = re.sub(r'\s(?:ac|ri):[A-Za-z0-9_-]+="[^"]*"', "", storage_html)
    cleaned = re.sub(r'\sdata-[A-Za-z0-9_-]+="[^"]*"', "", cleaned)
    cleaned = re.sub(
        r'<span class="placeholder-inline-tasks">\s*</span>',
        "",
        cleaned,
        flags=re.IGNORECASE,
    )
    cleaned = re.sub(
        r"\b[0-9]+\s+[0-9a-f]{8}-[0-9a-f-]{27}\s+incomplete\b",
        "",
        cleaned,
        flags=re.IGNORECASE,
    )
    return cleaned


def _projection_is_lossy(markdown: str) -> bool:
    return any(
        marker in markdown
        for marker in (
            "<table",
            "</table>",
            "<a href=",
            "Untitled Diagram-",
            "placeholder-inline-tasks",
        )
    )


def render_storage_to_markdown(storage_html: str) -> str:
    storage_html = _sanitize_storage_html_for_markdown(storage_html)
    try:
        proc = subprocess.run(
            ["pandoc", "-f", "html", "-t", "gfm", "--wrap=none"],
            input=storage_html,
            text=True,
            capture_output=True,
            check=True,
        )
    except FileNotFoundError as exc:
        raise ToolError("pandoc is required for HTML-to-markdown conversion") from exc
    except subprocess.CalledProcessError as exc:
        raise ToolError(f"pandoc failed: {exc.stderr.strip()}") from exc
    return _clean_markdown_projection(proc.stdout)


def render_page_markdown(page: dict[str, Any], session: Session) -> str:
    page_id = str(page.get("id") or "")
    title = str(page.get("title") or "(untitled)")
    page_url = ""
    if session.base_url and page_id:
        page_url = f"{session.base_url.rstrip('/')}/wiki/pages/viewpage.action?pageId={page_id}"
    created = str(page.get("createdAt") or "")
    version = page.get("version")
    if not isinstance(version, dict):
        version = {}
    updated = str(version.get("createdAt") or "")
    body = render_storage_to_markdown(storage_value(page))
    lines = [f"# {title}", ""]
    if page_url:
        lines.append(f"- URL: {page_url}")
    if page_id:
        lines.append(f"- Page ID: {page_id}")
    if page.get("spaceId"):
        lines.append(f"- Space ID: {page.get('spaceId')}")
    if created:
        lines.append(f"- Created: {created}")
    if updated:
        lines.append(f"- Updated: {updated}")
    if version.get("number") is not None:
        lines.append(f"- Version: {version.get('number')}")
    if _projection_is_lossy(body):
        lines.append(
            f"- Projection: approximate markdown; use `gotta confluence get {page_id} --output body` "
            "for canonical Confluence storage HTML when page layout, embedded diagrams, tables, "
            "or macros matter"
        )
    lines.extend(["", "---", "", body.rstrip(), ""])
    return "\n".join(lines)


def render_comment_markdown(comment: dict[str, Any], session: Session) -> str:
    comment_id = str(comment.get("id") or "")
    page_id = str(comment.get("pageId") or "")
    title = str(comment.get("title") or "Confluence Comment")
    comment_url = ""
    if session.base_url and page_id and comment_id:
        comment_url = (
            f"{session.base_url.rstrip('/')}/wiki/pages/viewpage.action?"
            f"pageId={page_id}&focusedCommentId={comment_id}"
        )
    created = str(comment.get("createdAt") or "")
    version = comment.get("version")
    if not isinstance(version, dict):
        version = {}
    updated = str(version.get("createdAt") or "")
    body = render_storage_to_markdown(comment_storage_value(comment))
    lines = [f"# {title}", ""]
    if comment_url:
        lines.append(f"- URL: {comment_url}")
    if comment_id:
        lines.append(f"- Comment ID: {comment_id}")
    if page_id:
        lines.append(f"- Page ID: {page_id}")
    if created:
        lines.append(f"- Created: {created}")
    if updated:
        lines.append(f"- Updated: {updated}")
    if version.get("number") is not None:
        lines.append(f"- Version: {version.get('number')}")
    lines.extend(["", "---", "", body.rstrip(), ""])
    return "\n".join(lines)


def format_resolve_page_payload(
    *,
    session: Session,
    parent_page: dict[str, Any],
    title: str,
    matches: list[dict[str, Any]],
) -> dict[str, Any]:
    parent_id = str(parent_page.get("id") or "")
    payload = {
        "title": title,
        "parent": normalize_page_summary(parent_page, session),
        "matchCount": len(matches),
        "found": len(matches) == 1,
        "matches": [normalize_page_summary(match, session) for match in matches],
    }
    payload["target"] = {
        "parentId": parent_id,
        "spaceId": str(parent_page.get("spaceId") or ""),
        "title": title,
    }
    return payload


def render_resolve_page_summary(payload: dict[str, Any]) -> str:
    parent = payload.get("parent")
    if not isinstance(parent, dict):
        parent = {}
    title = str(payload.get("title") or "")
    lines = [
        f"title: {title}",
        f"parent: {parent.get('title') or '(untitled)'} ({parent.get('id') or 'unknown'})",
        f"space_id: {payload.get('target', {}).get('spaceId') if isinstance(payload.get('target'), dict) else ''}",
        f"match_count: {payload.get('matchCount') or 0}",
    ]
    matches = payload.get("matches")
    if not isinstance(matches, list) or not matches:
        lines.append("matches: none")
        return "\n".join(lines) + "\n"
    lines.append("matches:")
    for page in matches:
        if not isinstance(page, dict):
            continue
        page_id = str(page.get("id") or "")
        title = str(page.get("title") or "(untitled)")
        url = str(page.get("url") or "")
        line = f"- {title}"
        details = [f"locator `confluence:{page_id}`"] if page_id else []
        if url:
            details.append(url)
        if details:
            line += " - " + ", ".join(details)
        lines.append(line)
    return "\n".join(lines) + "\n"


def format_create_page_preview(
    *,
    session: Session,
    parent_page: dict[str, Any],
    title: str,
    body_html: str,
    body_source: str,
    sibling_matches: list[dict[str, Any]],
) -> dict[str, Any]:
    parent_id = str(parent_page.get("id") or "")
    space_id = str(parent_page.get("spaceId") or "")
    return {
        "mode": "dry-run",
        "bodySource": body_source,
        "title": title,
        "parent": normalize_page_summary(parent_page, session),
        "target": {
            "apiUrl": pages_collection_url(session),
            "parentId": parent_id,
            "spaceId": space_id,
            "title": title,
        },
        "siblingExists": bool(sibling_matches),
        "siblingCount": len(sibling_matches),
        "siblings": [normalize_page_summary(match, session) for match in sibling_matches],
        "bodyPreview": summarize_storage_html(body_html),
    }


def render_create_page_preview(payload: dict[str, Any]) -> str:
    parent = payload.get("parent")
    if not isinstance(parent, dict):
        parent = {}
    body_preview = payload.get("bodyPreview")
    if not isinstance(body_preview, dict):
        body_preview = {}
    target = payload.get("target")
    if not isinstance(target, dict):
        target = {}
    lines = [
        f"mode: {payload.get('mode') or 'dry-run'}",
        f"title: {payload.get('title') or ''}",
        f"parent: {parent.get('title') or '(untitled)'} ({parent.get('id') or 'unknown'})",
        f"space_id: {target.get('spaceId') or ''}",
        f"body_source: {payload.get('bodySource') or ''}",
        f"sibling_exists: {'yes' if payload.get('siblingExists') else 'no'}",
        f"sibling_count: {payload.get('siblingCount') or 0}",
        f"create_target: {target.get('apiUrl') or ''}",
        f"body_chars: {body_preview.get('chars') or 0}",
    ]
    preview = str(body_preview.get("preview") or "").strip()
    if preview:
        lines.append(f"body_preview: {preview}")
    siblings = payload.get("siblings")
    if isinstance(siblings, list) and siblings:
        lines.append("siblings:")
        for page in siblings:
            if not isinstance(page, dict):
                continue
            page_id = str(page.get("id") or "")
            url = str(page.get("url") or "")
            line = f"- {page.get('title') or '(untitled)'}"
            details = [f"locator `confluence:{page_id}`"] if page_id else []
            if url:
                details.append(url)
            if details:
                line += " - " + ", ".join(details)
            lines.append(line)
    return "\n".join(lines) + "\n"


def format_created_page_result(page: dict[str, Any], session: Session) -> dict[str, Any]:
    result = normalize_page_summary(page, session)
    result["mode"] = "created"
    return result


def load_create_page_body(args: argparse.Namespace) -> tuple[str, str]:
    if args.from_markdown:
        markdown = read_text(Path(args.from_markdown))
        markdown = strip_matching_leading_h1(markdown, title=args.title)
        return render_markdown_to_storage(markdown), "markdown"
    if args.from_html:
        return read_text(Path(args.from_html)), "html"
    raise ToolError("missing create-page body input")


def cmd_resolve_page(args: argparse.Namespace) -> int:
    session, parent_page = fetch_page(parse_page_ref(args.parent))
    matches = find_child_pages_by_title(
        session,
        parent_page=parent_page,
        title=args.title,
    )
    payload = format_resolve_page_payload(
        session=session,
        parent_page=parent_page,
        title=args.title,
        matches=matches,
    )
    if args.output == "json":
        print_json(payload)
        return 0
    sys.stdout.write(render_resolve_page_summary(payload))
    return 0


def cmd_create_page(args: argparse.Namespace) -> int:
    body_html, body_source = load_create_page_body(args)
    session, parent_page = fetch_page(parse_page_ref(args.parent))
    siblings = find_child_pages_by_title(
        session,
        parent_page=parent_page,
        title=args.title,
    )
    if not args.apply:
        payload = format_create_page_preview(
            session=session,
            parent_page=parent_page,
            title=args.title,
            body_html=body_html,
            body_source=body_source,
            sibling_matches=siblings,
        )
        if args.output == "json":
            print_json(payload)
            return 0
        sys.stdout.write(render_create_page_preview(payload))
        return 0
    if siblings:
        sibling = normalize_page_summary(siblings[0], session)
        raise ToolError(
            "a sibling page with the same title already exists under the requested parent: "
            f"confluence:{sibling['id']} {sibling.get('url') or ''}".strip()
        )
    created = create_page(
        session,
        parent_page=parent_page,
        title=args.title,
        body_html=body_html,
    )
    payload = format_created_page_result(created, session)
    if args.output == "json":
        print_json(payload)
        return 0
    lines = [
        "mode: created",
        f"title: {payload.get('title') or ''}",
        f"page: {payload.get('id') or ''}",
    ]
    if payload.get("spaceId"):
        lines.append(f"space_id: {payload['spaceId']}")
    if payload.get("parentId"):
        lines.append(f"parent_id: {payload['parentId']}")
    if payload.get("url"):
        lines.append(f"url: {payload['url']}")
    sys.stdout.write("\n".join(lines) + "\n")
    return 0


def cmd_get(args: argparse.Namespace) -> int:
    session, content_kind, content = fetch_read_target(parse_content_ref(args.page))
    if args.output == "body":
        if content_kind == "comment":
            sys.stdout.write(comment_storage_value(content))
        else:
            sys.stdout.write(storage_value(content))
        return 0
    if args.output == "markdown":
        if content_kind == "comment":
            sys.stdout.write(render_comment_markdown(content, session))
        else:
            sys.stdout.write(render_page_markdown(content, session))
        return 0
    if args.output == "meta":
        if content_kind == "comment":
            content = {
                "id": content["id"],
                "type": "comment",
                "pageId": content.get("pageId"),
                "version": content.get("version"),
                "status": content.get("status"),
            }
        else:
            content = {
                "id": content["id"],
                "title": content["title"],
                "version": content["version"],
                "status": content.get("status"),
                "spaceId": content.get("spaceId"),
            }
    print_json(content)
    return 0


def cmd_find(args: argparse.Namespace) -> int:
    needle = read_single_input(args.needle, args.needle_file, input_name="needle input")
    _, page = fetch_page(parse_page_ref(args.page))
    print_json(
        count_with_context(
            storage_value(page),
            needle,
            context_chars=DEFAULT_FIND_CONTEXT_CHARS,
        )
    )
    return 0


def cmd_search(args: argparse.Namespace) -> int:
    base_url = args.base_url.strip()
    session = load_session(PageRef(base_url=base_url))
    result = search_confluence(
        session,
        cql=build_text_search_cql(args),
        limit=args.limit,
        cursor=args.next,
    )
    if args.output == "json":
        print_json(result)
        return 0
    sys.stdout.write(render_search_markdown(result))
    return 0


def cmd_cql(args: argparse.Namespace) -> int:
    base_url = args.base_url.strip()
    session = load_session(PageRef(base_url=base_url))
    result = search_confluence(
        session,
        cql=args.query,
        limit=args.limit,
        cursor=args.next,
    )
    if args.output == "json":
        print_json(result)
        return 0
    sys.stdout.write(render_search_markdown(result))
    return 0


def cmd_replace(args: argparse.Namespace) -> int:
    old = read_explicit_input(args.old_inline, args.from_file, input_name="--from/--from-file")
    new = read_explicit_input(args.new_inline, args.to_file, input_name="--to/--to-file")
    session, page = fetch_page(parse_page_ref(args.page))
    ensure_base_version(page, args.base_version)
    before = storage_value(page)
    after = replace_exact(before, old, new, expected_count=args.expected_count)
    if not args.apply:
        diff = unified_diff(before, after)
        sys.stdout.write(diff + ("\n" if diff else ""))
        return 0
    if after == before:
        print_json(format_noop_result(page))
        return 0
    updated = update_page(session, page, after, message=args.message)
    print_json(format_update_result(updated))
    return 0


def cmd_replace_section(args: argparse.Namespace) -> int:
    replacement = read_explicit_input(
        args.replacement, args.replacement_file, input_name="--replacement/--replacement-file"
    )
    session, page = fetch_page(parse_page_ref(args.page))
    ensure_base_version(page, args.base_version)
    before = storage_value(page)
    after = replace_between(
        before,
        args.start,
        args.end,
        replacement,
        include_markers=args.include_markers,
    )
    if not args.apply:
        diff = unified_diff(before, after)
        sys.stdout.write(diff + ("\n" if diff else ""))
        return 0
    if after == before:
        print_json(format_noop_result(page))
        return 0
    updated = update_page(session, page, after, message=args.message)
    print_json(format_update_result(updated))
    return 0


def cmd_render_markdown(args: argparse.Namespace) -> int:
    markdown = read_single_input(args.markdown, args.file, input_name="markdown input")
    sys.stdout.write(render_markdown_to_storage(markdown))
    return 0


def cmd_update_body(args: argparse.Namespace) -> int:
    new_body = read_single_input(
        args.body_inline,
        args.from_file,
        input_name="--from-file, inline body, or stdin",
    )
    session, page = fetch_page(parse_page_ref(args.page))
    ensure_base_version(page, args.base_version)
    before = storage_value(page)
    after = new_body
    if not args.apply:
        diff = unified_diff(before, after)
        sys.stdout.write(diff + ("\n" if diff else ""))
        return 0
    if after == before:
        print_json(format_noop_result(page))
        return 0
    updated = update_page(
        session,
        page,
        after,
        message=args.message,
    )
    print_json(format_update_result(updated))
    return 0


def cmd_batch(args: argparse.Namespace) -> int:
    operations = load_operations(args.operations_file)
    session, page = fetch_page(parse_page_ref(args.page))
    ensure_base_version(page, args.base_version)
    before = storage_value(page)
    after = before
    for index, operation in enumerate(operations, start=1):
        after = apply_operation(after, operation, index=index)
    if not args.apply:
        diff = unified_diff(before, after)
        sys.stdout.write(diff + ("\n" if diff else ""))
        return 0
    if after == before:
        result = format_noop_result(page)
        result["operations"] = len(operations)
        print_json(result)
        return 0
    updated = update_page(session, page, after, message=args.message)
    result = format_update_result(updated)
    result["operations"] = len(operations)
    print_json(result)
    return 0


def cmd_auth(args: argparse.Namespace) -> int:
    base_url = args.base_url.strip()
    if args.full:
        oauth_state = run_oauth_bootstrap(base_url=base_url)
        cloud_id = oauth_state.get("cloud_id")
        expires_at = oauth_state.get("expires_at")
        selected_site_url = str(oauth_state.get("base_url") or site_root(base_url)).strip()
    else:
        session = load_session(PageRef(base_url=site_root(base_url)))
        status = atlassian_status_payload(
            base_url=base_url,
            auth_command="confluence",
        )
        cloud_id = session.cloud_id
        expires_at = status.get("expiresAt")
        selected_site_url = str(session.base_url or status.get("baseUrl") or site_root(base_url)).strip()
    selected_confluence_url = f"{selected_site_url}/wiki" if selected_site_url else ""
    if selected_confluence_url:
        persist_selected_base_urls(selected_confluence_url)
    print_json(
        {
            "authenticated": True,
            "cloud_id": cloud_id,
            "expires_at": expires_at,
            "base_url": selected_confluence_url,
            "token_file": str(atl.TOKEN_FILE),
            "cloud_id_file": str(atl.CLOUD_ID_FILE),
            "oauth_dir": str(atl.OAUTH_DIR),
        }
    )
    return 0


def cmd_status(args: argparse.Namespace) -> int:
    payload = atlassian_status_payload(
        base_url=args.base_url,
        check_token=args.check,
        auth_command="confluence",
    )
    payload["surface"] = "confluence"
    if payload.get("baseUrl"):
        payload["confluenceBaseUrl"] = str(payload["baseUrl"]).rstrip("/") + "/wiki"
    if args.output == "json":
        print_json(payload)
        return 0
    lines = [
        "surface\tconfluence",
        f"base_url\t{payload.get('confluenceBaseUrl') or ''}",
        f"credentials_configured\t{str(bool(payload.get('credentialsConfigured'))).lower()}",
        f"session_status\t{payload.get('sessionStatus') or 'missing'}",
        f"expires_at\t{payload.get('expiresAt') or ''}",
        f"has_refresh_token\t{str(bool(payload.get('hasRefreshToken'))).lower()}",
        f"token_file\t{payload.get('tokenFile') or ''}",
        f"next_step\t{payload.get('nextStep') or ''}",
    ]
    if payload.get("tokenPreflight"):
        lines.append(f"token_preflight\t{payload['tokenPreflight']}")
    sys.stdout.write("\n".join(lines) + "\n")
    return 0


def resolve_mcp_runtime(
    base_url: str,
) -> tuple[str, str, str, str, str, str, str, str]:
    config_env = load_atlassian_config_env()
    client_id, client_secret, redirect_uri, scope = load_oauth_runtime_config()
    confluence_url = base_url.strip().rstrip("/")
    if not confluence_url:
        raise ToolError("missing Confluence base URL")
    site_url = site_root(confluence_url)
    toolsets = (
        os.environ.get("GOTTA_ATLASSIAN_TOOLSETS", "").strip()
        or config_env.get("GOTTA_ATLASSIAN_TOOLSETS", "").strip()
        or "all"
    )
    session = load_session(PageRef(base_url=site_url))
    return (
        client_id,
        client_secret,
        redirect_uri,
        scope,
        site_url,
        confluence_url,
        toolsets,
        session.cloud_id,
    )


def validate_mcp_passthrough_args(args: list[str]) -> None:
    for arg in args:
        flag = arg.split("=", 1)[0]
        if flag in DISALLOWED_MCP_PASSTHROUGH_FLAGS:
            raise ToolError(
                f"{flag} is managed by gotta confluence mcp; configure durable OAuth app "
                f"settings via env or {atl.provider_env_reference('atlassian')} instead"
            )


def is_mcp_metadata_only_invocation(args: list[str]) -> bool:
    return any(arg in {"--help", "--version", "-h"} for arg in args)


def cmd_mcp(args: argparse.Namespace) -> int:
    mcp_atlassian = shutil.which("mcp-atlassian")
    if not mcp_atlassian:
        raise ToolError(
            "missing required command: mcp-atlassian. "
            "Likely next step: install the Atlassian MCP server CLI"
        )

    passthrough = list(args.mcp_args)
    if passthrough and passthrough[0] == "--":
        passthrough = passthrough[1:]
    validate_mcp_passthrough_args(passthrough)

    env = os.environ.copy()
    env["FASTMCP_SHOW_CLI_BANNER"] = "false"
    env["FASTMCP_CHECK_FOR_UPDATES"] = "off"
    # Force mcp-atlassian to use the file-backed gotta auth cache.
    # Its keyring path can drift stale independently and then break token refresh.
    env.setdefault("PYTHON_KEYRING_BACKEND", "keyring.backends.fail.Keyring")

    if is_mcp_metadata_only_invocation(passthrough):
        os.execvpe(mcp_atlassian, [mcp_atlassian, *passthrough], env)
        return 0

    try:
        (
            client_id,
            client_secret,
            redirect_uri,
            scope,
            site_url,
            confluence_url,
            toolsets,
            cloud_id,
        ) = resolve_mcp_runtime(args.base_url)
    except ToolError as exc:
        raise ToolError(
            f"{exc}. Likely next step: run 'gotta confluence auth'"
        ) from exc

    argv = [
        mcp_atlassian,
        "--transport",
        "stdio",
        "--jira-url",
        site_url,
        "--confluence-url",
        confluence_url,
        "--toolsets",
        toolsets,
        "--oauth-client-id",
        client_id,
        "--oauth-client-secret",
        client_secret,
        "--oauth-redirect-uri",
        redirect_uri,
        "--oauth-scope",
        scope,
        "--oauth-cloud-id",
        cloud_id,
        *passthrough,
    ]
    os.execvpe(mcp_atlassian, argv, env)
    return 0


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="gotta confluence",
        description=(
            "Standalone Confluence page-body fetch/find/edit toolbelt. "
            "Interactive commands may open browser reauthorization if the "
            "cached Atlassian session is invalid."
        ),
    )
    sub = parser.add_subparsers(dest="command", required=True)

    p = sub.add_parser("status", help="inspect local Confluence/Atlassian auth readiness")
    p.add_argument(
        "--base-url",
        default=default_base_url(),
        help=f"Confluence base URL override (default: {CONFLUENCE_BASE_URL_ENV})",
    )
    p.add_argument("--check", action="store_true", help="run a token preflight against Atlassian")
    p.add_argument("--output", choices=["json", "summary"], default="summary")
    p.set_defaults(func=cmd_status)

    p = sub.add_parser("get", help="fetch page body HTML and metadata")
    p.add_argument("page", help=PAGE_ARG_HELP)
    p.add_argument(
        "--output",
        choices=["json", "body", "meta", "markdown"],
        default="markdown",
        help="output mode; defaults to markdown for first-pass reading. use body for canonical "
        "Confluence storage HTML when page layout, tasks, or diagrams need highest fidelity",
    )
    p.set_defaults(func=cmd_get)

    p = sub.add_parser(
        "find",
        help="count exact fragments in page body HTML and return surrounding context",
    )
    p.add_argument("page", help=PAGE_ARG_HELP)
    p.add_argument("needle", nargs="?", help="exact fragment; reads stdin if omitted")
    p.add_argument("--needle-file", help="file containing exact fragment")
    p.set_defaults(func=cmd_find)

    p = sub.add_parser(
        "search",
        help="search Confluence content with plain text",
    )
    p.add_argument("query", help="plain-text search query")
    p.add_argument(
        "--title-only",
        action="store_true",
        help="search only page titles instead of full text",
    )
    p.add_argument("--space", help="restrict results to a Confluence space key")
    p.add_argument(
        "--type",
        choices=["page", "blogpost", "attachment", "comment", "all"],
        default="page",
        help="restrict results to one Confluence content type; defaults to page-first search",
    )
    p.add_argument("--limit", type=int, default=10, help="maximum results to request")
    p.add_argument("--next", help="continuation token from a previous search response")
    p.add_argument("--output", choices=["markdown", "json"], default="markdown")
    p.add_argument(
        "--base-url",
        default=default_base_url(),
        help=f"Confluence base URL override (default: {CONFLUENCE_BASE_URL_ENV})",
    )
    p.set_defaults(func=cmd_search)

    p = sub.add_parser("cql", help="search Confluence content with raw CQL")
    p.add_argument("query", help="raw CQL search expression")
    p.add_argument("--limit", type=int, default=10, help="maximum results to request")
    p.add_argument("--next", help="continuation token from a previous search response")
    p.add_argument("--output", choices=["markdown", "json"], default="markdown")
    p.add_argument(
        "--base-url",
        default=default_base_url(),
        help=f"Confluence base URL override (default: {CONFLUENCE_BASE_URL_ENV})",
    )
    p.set_defaults(func=cmd_cql)

    p = sub.add_parser(
        "resolve-page",
        help="resolve an exact child page title under a specific parent page",
    )
    p.add_argument("--parent", required=True, help=PAGE_ARG_HELP)
    p.add_argument("--title", required=True, help="exact child page title to resolve")
    p.add_argument("--output", choices=["summary", "json"], default="summary")
    p.set_defaults(func=cmd_resolve_page)

    p = sub.add_parser(
        "create-page",
        help="preview or create a child page under a specific parent page",
    )
    p.add_argument("--parent", required=True, help=PAGE_ARG_HELP)
    p.add_argument("--title", required=True, help="final child page title")
    body_group = p.add_mutually_exclusive_group(required=True)
    body_group.add_argument(
        "--from-markdown",
        help=(
            "markdown file to render to Confluence storage HTML before create; "
            "a leading H1 that matches --title is stripped before render"
        ),
    )
    body_group.add_argument(
        "--from-html",
        help="file containing Confluence storage HTML to create verbatim",
    )
    mode_group = p.add_mutually_exclusive_group()
    mode_group.add_argument(
        "--apply",
        action="store_true",
        help="create the page; default is dry-run preview",
    )
    mode_group.add_argument(
        "--dry-run",
        action="store_true",
        help="preview the create target explicitly; also the default when --apply is omitted",
    )
    p.add_argument("--output", choices=["summary", "json"], default="summary")
    p.set_defaults(func=cmd_create_page)

    p = sub.add_parser("replace", help="replace an exact fragment in page body HTML")
    p.add_argument("page", help=PAGE_ARG_HELP)
    p.add_argument("--from", dest="old_inline", help="old exact fragment")
    p.add_argument("--from-file", help="file containing old exact fragment")
    p.add_argument("--to", dest="new_inline", help="new replacement fragment")
    p.add_argument("--to-file", help="file containing new replacement fragment")
    p.add_argument(
        "--expected-count",
        type=int,
        default=1,
        help="require this exact number of matches before replacing",
    )
    p.add_argument(
        "--base-version",
        type=int,
        help="require the current page version to match before diffing or writing",
    )
    p.add_argument("--message", default="Replace Confluence body fragment")
    p.add_argument(
        "--apply",
        action="store_true",
        help="write one full-body update as a requested minor edit; default is dry-run diff",
    )
    p.set_defaults(func=cmd_replace)

    p = sub.add_parser(
        "replace-section", help="replace content between exact start/end markers"
    )
    p.add_argument("page", help=PAGE_ARG_HELP)
    p.add_argument("--start", required=True, help="exact start marker")
    p.add_argument("--end", required=True, help="exact end marker")
    p.add_argument("--replacement", help="replacement content")
    p.add_argument("--replacement-file", help="file containing replacement content")
    p.add_argument(
        "--include-markers",
        action="store_true",
        help="replace the markers themselves instead of only the content between them",
    )
    p.add_argument(
        "--base-version",
        type=int,
        help="require the current page version to match before diffing or writing",
    )
    p.add_argument("--message", default="Replace Confluence body section")
    p.add_argument(
        "--apply",
        action="store_true",
        help="write one full-body update as a requested minor edit; default is dry-run diff",
    )
    p.set_defaults(func=cmd_replace_section)

    p = sub.add_parser(
        "render-markdown",
        help="convert markdown to Confluence storage HTML for safe insertion",
    )
    p.add_argument("markdown", nargs="?", help="inline markdown; reads stdin if omitted")
    p.add_argument("--file", help="file containing markdown")
    p.set_defaults(func=cmd_render_markdown)

    p = sub.add_parser(
        "update-body",
        help="diff or write a full replacement page body from file, inline input, or stdin",
    )
    p.add_argument("page", help=PAGE_ARG_HELP)
    p.add_argument("body_inline", nargs="?", help="inline replacement body HTML")
    p.add_argument("--from-file", help="file containing replacement body HTML")
    p.add_argument(
        "--base-version",
        type=int,
        help="require the current page version to match before diffing or writing",
    )
    p.add_argument("--message", default="Update Confluence page body")
    p.add_argument(
        "--apply",
        action="store_true",
        help="write one full-body update as a requested minor edit; default is dry-run diff",
    )
    p.set_defaults(func=cmd_update_body)

    p = sub.add_parser(
        "batch",
        help="apply multiple exact/section replacements in memory, then write one full-body update",
    )
    p.add_argument("page", help=PAGE_ARG_HELP)
    p.add_argument(
        "--operations-file",
        required=True,
        help="JSON array describing replace/replace-section operations",
    )
    p.add_argument(
        "--base-version",
        type=int,
        help="require the current page version to match before diffing or writing",
    )
    p.add_argument("--message", default="Apply Confluence body batch")
    p.add_argument(
        "--apply",
        action="store_true",
        help="write one full-body update as a requested minor edit; default is dry-run diff",
    )
    p.set_defaults(func=cmd_batch)

    p = sub.add_parser(
        "auth",
        help="ensure local Atlassian OAuth state is usable; refresh silently when possible",
    )
    p.add_argument(
        "--base-url",
        default=default_base_url(),
        help=f"Confluence base URL override (default: {CONFLUENCE_BASE_URL_ENV})",
    )
    p.add_argument(
        "--full",
        action="store_true",
        help="force a full browser OAuth bootstrap instead of reusing or refreshing cached state",
    )
    p.set_defaults(func=cmd_auth)

    p = sub.add_parser(
        "mcp",
        help="start mcp-atlassian with the same managed local OAuth/session contract",
    )
    p.add_argument(
        "--base-url",
        default=default_base_url(),
        help=f"Confluence base URL override for MCP startup (default: {CONFLUENCE_BASE_URL_ENV})",
    )
    p.add_argument(
        "mcp_args",
        nargs=argparse.REMAINDER,
        help="raw mcp-atlassian flags after --",
    )
    p.set_defaults(func=cmd_mcp)

    return parser


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    if is_long_help_request(argv):
        return print_long_help(parser)
    args = parser.parse_args(argv)
    return int(args.func(args))


if __name__ == "__main__":
    if hasattr(signal, "SIGPIPE"):
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    try:
        raise SystemExit(main(sys.argv[1:]))
    except BrokenPipeError:
        raise SystemExit(0)
    except ToolError as exc:
        raise SystemExit(die(str(exc)))



"""Shared Google OAuth and Drive helpers."""

from __future__ import annotations

import http.server
import json
import os
from pathlib import Path
import secrets
import socketserver
import sys
import threading
import time
import urllib.error
import urllib.parse
import urllib.request
import webbrowser
from typing import Any

from gotta.compat import tomllib
from gotta.config import (
    config_file_candidates,
    env_or_config,
    extract_provider_env,
    primary_config_file,
    provider_env_reference,
    user_state_dir,
)

OAUTH_DIR = user_state_dir() / "auth" / "google"
TOKEN_FILE = OAUTH_DIR / "oauth.json"
CONFIG_FILE = primary_config_file()
GOOGLE_AUTHORIZE_URL = "https://accounts.google.com/o/oauth2/v2/auth"
GOOGLE_TOKEN_URL = "https://oauth2.googleapis.com/token"
GOOGLE_DOCS_API_URL = "https://docs.googleapis.com/v1/documents"
GOOGLE_DRIVE_API_URL = "https://www.googleapis.com/drive/v3/files"
GOOGLE_SHEETS_API_URL = "https://sheets.googleapis.com/v4/spreadsheets"
GOOGLE_DOC_MIME = "application/vnd.google-apps.document"
GOOGLE_SHEET_MIME = "application/vnd.google-apps.spreadsheet"
DEFAULT_REDIRECT_URI = "http://localhost:8091/callback"
DEFAULT_SCOPE = (
    "https://www.googleapis.com/auth/documents.readonly "
    "https://www.googleapis.com/auth/drive.readonly "
    "https://www.googleapis.com/auth/spreadsheets.readonly"
)
DEFAULT_DRIVE_FIELDS = (
    "id,name,mimeType,createdTime,modifiedTime,webViewLink,size,"
    "owners(displayName,emailAddress)"
)


class GoogleError(RuntimeError):
    """Raised when the Google surface cannot satisfy a request."""


def parse_toml_file(path: Path) -> dict[str, Any]:
    try:
        with path.open("rb") as handle:
            data = tomllib.load(handle)
    except tomllib.TOMLDecodeError as exc:
        raise GoogleError(f"invalid TOML config {path}: {exc}") from exc
    if not isinstance(data, dict):
        raise GoogleError(f"invalid TOML config {path}: expected table")
    return data


def load_provider_config_env(*provider_names: str) -> dict[str, str]:
    for config_file in config_file_candidates():
        if not config_file.exists():
            continue
        config = parse_toml_file(config_file)
        for provider_name in provider_names:
            result = extract_provider_env(config, provider_name)
            if result:
                return result
    return {}


def load_google_config_env() -> dict[str, str]:
    return load_provider_config_env("google")


def google_credentials_present() -> bool:
    config_env = load_google_config_env()
    client_id = env_or_config(
        config_env,
        "GOTTA_GOOGLE_OAUTH_CLIENT_ID",
    )
    client_secret = env_or_config(
        config_env,
        "GOTTA_GOOGLE_OAUTH_CLIENT_SECRET",
    )
    return bool(client_id and client_secret)


def load_oauth_runtime_config() -> tuple[str, str, str, str]:
    config_env = load_google_config_env()
    client_id = env_or_config(
        config_env,
        "GOTTA_GOOGLE_OAUTH_CLIENT_ID",
    )
    client_secret = env_or_config(
        config_env,
        "GOTTA_GOOGLE_OAUTH_CLIENT_SECRET",
    )
    redirect_uri = env_or_config(
        config_env,
        "GOTTA_GOOGLE_OAUTH_REDIRECT_URI",
        default=DEFAULT_REDIRECT_URI,
    )
    scope = env_or_config(
        config_env,
        "GOTTA_GOOGLE_OAUTH_SCOPE",
        default=DEFAULT_SCOPE,
    )
    if not client_id or not client_secret:
        raise GoogleError(
            "missing Google OAuth client credentials; set "
            "GOTTA_GOOGLE_OAUTH_CLIENT_ID and GOTTA_GOOGLE_OAUTH_CLIENT_SECRET "
            "or configure them under "
            f"{provider_env_reference('google')}"
        )
    if not redirect_uri:
        raise GoogleError("missing Google OAuth redirect URI")
    if not scope:
        raise GoogleError("missing Google OAuth scope")
    return client_id, client_secret, redirect_uri, scope


def ensure_oauth_dir() -> None:
    OAUTH_DIR.mkdir(parents=True, exist_ok=True)
    os.chmod(OAUTH_DIR, 0o700)


def write_secret_text(path: Path, value: str) -> None:
    ensure_oauth_dir()
    fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o600)
    try:
        with os.fdopen(fd, "w", encoding="utf-8") as handle:
            handle.write(value)
    finally:
        try:
            os.chmod(path, 0o600)
        except FileNotFoundError:
            pass


def persist_oauth_state(oauth_state: dict[str, Any]) -> dict[str, Any]:
    write_secret_text(TOKEN_FILE, json.dumps(oauth_state))
    return oauth_state


def parse_json_bytes(data: bytes, *, context: str) -> dict[str, Any]:
    try:
        payload = json.loads(data)
    except json.JSONDecodeError as exc:
        raise GoogleError(f"invalid JSON from {context}: {exc}") from exc
    if not isinstance(payload, dict):
        raise GoogleError(f"unexpected JSON payload from {context}: expected object")
    return payload


def request_bytes(
    method: str,
    url: str,
    *,
    headers: dict[str, str] | None = None,
    data: bytes | None = None,
) -> bytes:
    req = urllib.request.Request(url, data=data, method=method)
    for key, value in (headers or {}).items():
        req.add_header(key, value)
    try:
        with urllib.request.urlopen(req) as response:
            return response.read()
    except urllib.error.HTTPError as exc:
        body = exc.read().decode("utf-8", errors="replace")
        raise GoogleError(f"{method} {url} failed with {exc.code}: {body}") from exc
    except urllib.error.URLError as exc:
        raise GoogleError(f"{method} {url} failed: {exc.reason}") from exc


def post_form_json(url: str, payload: dict[str, str], *, context: str) -> dict[str, Any]:
    encoded = urllib.parse.urlencode(payload).encode("utf-8")
    body = request_bytes(
        "POST",
        url,
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data=encoded,
    )
    return parse_json_bytes(body, context=context)


def authorization_url(client_id: str, redirect_uri: str, scope: str, state: str) -> str:
    params = urllib.parse.urlencode(
        {
            "client_id": client_id,
            "redirect_uri": redirect_uri,
            "response_type": "code",
            "scope": scope,
            "access_type": "offline",
            "include_granted_scopes": "true",
            "prompt": "consent",
            "state": state,
        }
    )
    return f"{GOOGLE_AUTHORIZE_URL}?{params}"


def start_oauth_callback_server(
    redirect_uri: str,
) -> tuple[socketserver.TCPServer, dict[str, Any]]:
    parsed = urllib.parse.urlparse(redirect_uri)
    hostname = parsed.hostname
    port = parsed.port or (443 if parsed.scheme == "https" else 80)
    if hostname not in {"localhost", "127.0.0.1"}:
        raise GoogleError(
            f"unsupported Google OAuth redirect host for local bootstrap: {hostname}"
        )

    callback_state: dict[str, Any] = {
        "code": None,
        "state": None,
        "error": None,
        "event": threading.Event(),
    }

    class ReusableTCPServer(socketserver.TCPServer):
        allow_reuse_address = True

    class CallbackHandler(http.server.BaseHTTPRequestHandler):
        def do_GET(self) -> None:  # noqa: N802
            params = urllib.parse.parse_qs(urllib.parse.urlparse(self.path).query)
            if "error" in params:
                callback_state["error"] = params["error"][0]
                callback_state["event"].set()
                self._send_response(
                    f"Authorization failed: {callback_state['error']}", status=400
                )
                return
            if "code" in params:
                callback_state["code"] = params["code"][0]
                callback_state["state"] = params.get("state", [""])[0]
                callback_state["event"].set()
                self._send_response("Authorization successful. You can close this window.")
                return
            self._send_response("Invalid callback: missing authorization code.", status=400)

        def _send_response(self, message: str, *, status: int = 200) -> None:
            self.send_response(status)
            self.send_header("Content-Type", "text/html")
            self.end_headers()
            self.wfile.write(
                (
                    "<!doctype html><html><body><p>"
                    + message
                    + "</p><script>window.close()</script></body></html>"
                ).encode("utf-8")
            )

        def log_message(self, _format: str, *args: Any) -> None:
            return

    try:
        server = ReusableTCPServer((hostname, port), CallbackHandler)
    except OSError as exc:
        raise GoogleError(
            f"failed to start Google OAuth callback server on {redirect_uri}: {exc}"
        ) from exc
    return server, callback_state


def exchange_authorization_code(
    client_id: str, client_secret: str, redirect_uri: str, code: str
) -> dict[str, Any]:
    return post_form_json(
        GOOGLE_TOKEN_URL,
        {
            "grant_type": "authorization_code",
            "client_id": client_id,
            "client_secret": client_secret,
            "redirect_uri": redirect_uri,
            "code": code,
        },
        context="Google OAuth token exchange",
    )


def run_oauth_bootstrap(*, interactive_ok: bool) -> dict[str, Any]:
    if not interactive_ok:
        raise GoogleError("Google OAuth bootstrap requires an interactive terminal")
    client_id, client_secret, redirect_uri, scope = load_oauth_runtime_config()
    state = secrets.token_urlsafe(16)
    auth_url = authorization_url(client_id, redirect_uri, scope, state)
    server, callback_state = start_oauth_callback_server(redirect_uri)
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    print("Google authorization is required; opening browser", file=sys.stderr)
    print(f"authorization URL: {auth_url}", file=sys.stderr)
    thread.start()
    try:
        if not webbrowser.open(auth_url):
            print(
                "failed to auto-open browser; open the authorization URL manually",
                file=sys.stderr,
            )
        if not callback_state["event"].wait(300):
            raise GoogleError("timed out waiting for Google OAuth callback")
        if callback_state["error"]:
            raise GoogleError(f"Google OAuth returned an error: {callback_state['error']}")
        if callback_state["state"] != state:
            raise GoogleError("OAuth state mismatch in callback")
        token_response = exchange_authorization_code(
            client_id,
            client_secret,
            redirect_uri,
            str(callback_state["code"] or ""),
        )
    finally:
        server.shutdown()
        server.server_close()

    access_token = str(token_response.get("access_token") or "").strip()
    refresh_token = str(token_response.get("refresh_token") or "").strip()
    if not access_token:
        raise GoogleError("Google OAuth bootstrap did not return an access token")
    expires_in = token_response.get("expires_in")
    expires_at = None
    if isinstance(expires_in, (int, float)):
        expires_at = time.time() + float(expires_in)
    oauth_state = {
        "access_token": access_token,
        "refresh_token": refresh_token,
        "scope": token_response.get("scope") or "",
        "token_type": token_response.get("token_type") or "Bearer",
        "expires_at": expires_at,
    }
    return persist_oauth_state(oauth_state)


def load_cached_oauth_state() -> dict[str, Any] | None:
    if not TOKEN_FILE.exists():
        return None
    try:
        data = json.loads(TOKEN_FILE.read_text(encoding="utf-8"))
    except json.JSONDecodeError as exc:
        raise GoogleError(f"invalid Google OAuth state file {TOKEN_FILE}: {exc}") from exc
    if not isinstance(data, dict):
        raise GoogleError(
            f"invalid Google OAuth state file {TOKEN_FILE}: expected JSON object"
        )
    return data


def google_status_payload() -> dict[str, Any]:
    config_env = load_google_config_env()
    payload: dict[str, Any] = {
        "oauthDir": str(OAUTH_DIR),
        "tokenFile": str(TOKEN_FILE),
        "tokenFileExists": TOKEN_FILE.exists(),
        "configFile": str(CONFIG_FILE),
        "configFileExists": CONFIG_FILE.exists(),
        "credentialsConfigured": google_credentials_present(),
        "credentialSources": {
            "envClientId": bool(
                env_or_config({}, "GOTTA_GOOGLE_OAUTH_CLIENT_ID")
            ),
            "envClientSecret": bool(
                env_or_config({}, "GOTTA_GOOGLE_OAUTH_CLIENT_SECRET")
            ),
            "configClientId": bool(
                env_or_config(config_env, "GOTTA_GOOGLE_OAUTH_CLIENT_ID")
            ),
            "configClientSecret": bool(
                env_or_config(config_env, "GOTTA_GOOGLE_OAUTH_CLIENT_SECRET")
            ),
        },
    }
    try:
        oauth_state = load_cached_oauth_state()
    except GoogleError as exc:
        payload["sessionStatus"] = "invalid"
        payload["error"] = str(exc)
        payload["nextStep"] = "run `gotta gdocs auth` or `gotta gdrive auth`"
        return payload
    if oauth_state is None:
        payload["sessionStatus"] = "missing"
        payload["nextStep"] = (
            "run `gotta gdocs auth` or `gotta gdrive auth` after configuring "
            f"{provider_env_reference('google')}"
            if payload["credentialsConfigured"]
            else f"configure Google OAuth credentials under {provider_env_reference('google')}"
        )
        return payload
    expires_at = oauth_state.get("expires_at")
    payload["sessionStatus"] = (
        "expired" if token_is_expired(oauth_state, skew_seconds=0) else "usable"
    )
    payload["expiresAt"] = expires_at
    payload["hasRefreshToken"] = bool(str(oauth_state.get("refresh_token") or "").strip())
    payload["scope"] = str(oauth_state.get("scope") or "")
    payload["tokenType"] = str(oauth_state.get("token_type") or "")
    if not payload["credentialsConfigured"]:
        payload["nextStep"] = (
            f"configure Google OAuth credentials under {provider_env_reference('google')}"
        )
    elif payload["sessionStatus"] in {"invalid", "missing"}:
        payload["nextStep"] = "run `gotta gdocs auth` or `gotta gdrive auth`"
    elif payload["sessionStatus"] == "expired":
        if payload["hasRefreshToken"]:
            payload["nextStep"] = (
                "rerun a native Google command first; gotta will usually refresh "
                "automatically when a refresh token is present. Re-run auth only "
                "if refresh does not recover the session"
            )
        else:
            payload["nextStep"] = "run `gotta gdocs auth` or `gotta gdrive auth`"
    else:
        payload["nextStep"] = "ready"
    return payload


def token_is_expired(oauth_state: dict[str, Any], *, skew_seconds: int = 300) -> bool:
    expires_at = oauth_state.get("expires_at")
    if not isinstance(expires_at, (int, float)):
        return True
    return time.time() + skew_seconds >= float(expires_at)


def refresh_oauth_state(oauth_state: dict[str, Any]) -> dict[str, Any]:
    refresh_token = str(oauth_state.get("refresh_token") or "").strip()
    if not refresh_token:
        raise GoogleError("cached Google OAuth state is missing a refresh token")
    client_id, client_secret, _, _ = load_oauth_runtime_config()
    response = post_form_json(
        GOOGLE_TOKEN_URL,
        {
            "grant_type": "refresh_token",
            "client_id": client_id,
            "client_secret": client_secret,
            "refresh_token": refresh_token,
        },
        context="Google OAuth token refresh",
    )
    access_token = str(response.get("access_token") or "").strip()
    if not access_token:
        raise GoogleError("Google OAuth refresh did not return an access token")
    expires_in = response.get("expires_in")
    expires_at = None
    if isinstance(expires_in, (int, float)):
        expires_at = time.time() + float(expires_in)
    oauth_state["access_token"] = access_token
    oauth_state["token_type"] = response.get("token_type") or oauth_state.get("token_type") or "Bearer"
    oauth_state["scope"] = response.get("scope") or oauth_state.get("scope") or ""
    oauth_state["expires_at"] = expires_at
    return persist_oauth_state(oauth_state)


def ensure_google_session(
    *,
    allow_bootstrap: bool,
    interactive_ok: bool,
    auth_command: str,
) -> dict[str, Any]:
    oauth_state = load_cached_oauth_state()
    if oauth_state is None:
        if allow_bootstrap and interactive_ok:
            return run_oauth_bootstrap(interactive_ok=interactive_ok)
        if google_credentials_present():
            raise GoogleError(
                f"no cached Google OAuth session is available; run 'gotta {auth_command} auth' "
                f"to create {TOKEN_FILE}"
            )
        raise GoogleError(
            f"missing Google OAuth credentials and session state; run 'gotta {auth_command} auth' "
            "after configuring GOTTA_GOOGLE_OAUTH_CLIENT_ID and GOTTA_GOOGLE_OAUTH_CLIENT_SECRET or "
            f"{provider_env_reference('google')}"
        )
    if token_is_expired(oauth_state):
        try:
            oauth_state = refresh_oauth_state(oauth_state)
        except GoogleError:
            if allow_bootstrap and interactive_ok:
                return run_oauth_bootstrap(interactive_ok=interactive_ok)
            raise
    return oauth_state


def google_json(url: str, access_token: str) -> dict[str, Any]:
    body = request_bytes(
        "GET",
        url,
        headers={"Authorization": f"Bearer {access_token}"},
    )
    return parse_json_bytes(body, context=url)


def google_bytes(url: str, access_token: str) -> bytes:
    return request_bytes(
        "GET",
        url,
        headers={"Authorization": f"Bearer {access_token}"},
    )


def parse_doc_ref(raw: str) -> tuple[str, str]:
    raw = raw.strip()
    parsed = urllib.parse.urlparse(raw)
    if parsed.scheme in {"http", "https"} and parsed.netloc == "docs.google.com":
        parts = [part for part in parsed.path.split("/") if part]
        if parts and parts[0] == "document" and "d" in parts:
            d_index = parts.index("d")
            if d_index + 1 < len(parts):
                doc_id = parts[d_index + 1]
                return doc_id, f"https://docs.google.com/document/d/{doc_id}/edit"
    if raw and "/" not in raw and " " not in raw:
        doc_id = raw
        return doc_id, f"https://docs.google.com/document/d/{doc_id}/edit"
    raise GoogleError(f"could not parse Google Docs document ID from input: {raw}")


def parse_drive_ref(raw: str) -> tuple[str, str]:
    raw = raw.strip()
    parsed = urllib.parse.urlparse(raw)
    if parsed.scheme in {"http", "https"}:
        if parsed.netloc == "drive.google.com":
            parts = [part for part in parsed.path.split("/") if part]
            if len(parts) >= 3 and parts[0] == "file" and parts[1] == "d":
                file_id = parts[2]
                return file_id, f"https://drive.google.com/file/d/{file_id}/view"
            query = urllib.parse.parse_qs(parsed.query)
            if "id" in query and query["id"]:
                file_id = query["id"][0]
                return file_id, f"https://drive.google.com/open?id={file_id}"
        if parsed.netloc == "docs.google.com":
            parts = [part for part in parsed.path.split("/") if part]
            if "d" in parts:
                d_index = parts.index("d")
                if d_index + 1 < len(parts):
                    file_id = parts[d_index + 1]
                    return file_id, raw
    if raw and "/" not in raw and " " not in raw:
        return raw, f"https://drive.google.com/open?id={raw}"
    raise GoogleError(f"could not parse Google Drive file ID from input: {raw}")


def parse_sheet_ref(raw: str) -> tuple[str, str]:
    raw = raw.strip()
    parsed = urllib.parse.urlparse(raw)
    if parsed.scheme in {"http", "https"} and parsed.netloc == "docs.google.com":
        parts = [part for part in parsed.path.split("/") if part]
        if parts and parts[0] == "spreadsheets" and "d" in parts:
            d_index = parts.index("d")
            if d_index + 1 < len(parts):
                sheet_id = parts[d_index + 1]
                return sheet_id, f"https://docs.google.com/spreadsheets/d/{sheet_id}/edit"
    if raw and "/" not in raw and " " not in raw:
        return raw, f"https://docs.google.com/spreadsheets/d/{raw}/edit"
    raise GoogleError(f"could not parse Google Sheets spreadsheet ID from input: {raw}")


def drive_file_meta(
    access_token: str,
    file_id: str,
    *,
    fields: str = DEFAULT_DRIVE_FIELDS,
) -> dict[str, Any]:
    url = (
        f"{GOOGLE_DRIVE_API_URL}/{urllib.parse.quote(file_id)}"
        f"?fields={urllib.parse.quote(fields, safe=',()')}"
    )
    return google_json(url, access_token)


def drive_export(access_token: str, file_id: str, mime_type: str) -> bytes:
    url = (
        f"{GOOGLE_DRIVE_API_URL}/{urllib.parse.quote(file_id)}/export"
        f"?mimeType={urllib.parse.quote(mime_type, safe='')}"
    )
    return google_bytes(url, access_token)


def drive_download_bytes(access_token: str, file_id: str) -> bytes:
    url = f"{GOOGLE_DRIVE_API_URL}/{urllib.parse.quote(file_id)}?alt=media"
    return google_bytes(url, access_token)


def escape_drive_query(value: str) -> str:
    return value.replace("\\", "\\\\").replace("'", "\\'")


def drive_search_files(access_token: str, q: str, *, limit: int, fields: str) -> list[dict[str, Any]]:
    params = urllib.parse.urlencode(
        {
            "q": q,
            "pageSize": str(limit),
            "orderBy": "modifiedTime desc",
            "fields": f"files({fields})",
        }
    )
    payload = google_json(f"{GOOGLE_DRIVE_API_URL}?{params}", access_token)
    files = payload.get("files")
    if not isinstance(files, list):
        raise GoogleError("unexpected Google Drive search response: missing files list")
    return [item for item in files if isinstance(item, dict)]


def sheets_spreadsheet_meta(access_token: str, spreadsheet_id: str) -> dict[str, Any]:
    url = (
        f"{GOOGLE_SHEETS_API_URL}/{urllib.parse.quote(spreadsheet_id)}"
        "?includeGridData=false&fields="
        "spreadsheetId,properties(title),sheets(properties(sheetId,title,index,gridProperties(rowCount,columnCount)))"
    )
    payload = google_json(url, access_token)
    payload["url"] = f"https://docs.google.com/spreadsheets/d/{spreadsheet_id}/edit"
    return payload


def sheets_values(
    access_token: str,
    spreadsheet_id: str,
    *,
    a1_range: str,
) -> dict[str, Any]:
    encoded_range = urllib.parse.quote(a1_range, safe="!':,")
    params = urllib.parse.urlencode(
        {
            "majorDimension": "ROWS",
            "valueRenderOption": "FORMATTED_VALUE",
            "dateTimeRenderOption": "FORMATTED_STRING",
        }
    )
    url = f"{GOOGLE_SHEETS_API_URL}/{urllib.parse.quote(spreadsheet_id)}/values/{encoded_range}?{params}"
    return google_json(url, access_token)
